import copy
from typing import Any, Optional

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.transforms as transforms
from segment_anything import SamAutomaticMaskGenerator
from segment_anything.modeling import MaskDecoder, PromptEncoder, TwoWayTransformer
from segment_anything.modeling.mask_decoder import MaskDecoder
from segment_anything.modeling.prompt_encoder import PromptEncoder
from segment_anything.utils.amg import build_all_layer_point_grids
from segment_anything.utils.transforms import ResizeLongestSide
from torchvision.transforms.functional import resize, to_pil_image

from efficientvit.models.efficientvit.backbone import EfficientViTBackbone, EfficientViTLargeBackbone
from efficientvit.models.nn import (
    ConvLayer,
    DAGBlock,
    FusedMBConv,
    IdentityLayer,
    MBConv,
    OpSequential,
    ResBlock,
    ResidualBlock,
    UpSampleLayer,
    build_norm,
)
from efficientvit.models.utils import build_kwargs_from_config, get_device

__all__ = [
    "SamPad",
    "SamResize",
    "SamNeck",
    "EfficientViTSamImageEncoder",
    "EfficientViTSam",
    "EfficientViTSamPredictor",
    "EfficientViTSamAutomaticMaskGenerator",
    "efficientvit_sam_l0",
    "efficientvit_sam_l1",
    "efficientvit_sam_l2",
    "efficientvit_sam_xl0",
    "efficientvit_sam_xl1",
]


class SamPad:
    def __init__(self, size: int, fill: float = 0, pad_mode="corner") -> None:
        self.size = size
        self.fill = fill
        self.pad_mode = pad_mode

    def __call__(self, image: torch.Tensor) -> torch.Tensor:
        h, w = image.shape[-2:]
        th, tw = self.size, self.size
        assert th >= h and tw >= w
        if self.pad_mode == "corner":
            image = F.pad(image, (0, tw - w, 0, th - h), value=self.fill)
        else:
            raise NotImplementedError
        return image

    def __repr__(self) -> str:
        return f"{type(self).__name__}(size={self.size},mode={self.pad_mode},fill={self.fill})"


class SamResize:
    def __init__(self, size: int) -> None:
        self.size = size

    def __call__(self, image: np.ndarray) -> np.ndarray:
        h, w, _ = image.shape
        long_side = max(h, w)
        if long_side != self.size:
            return self.apply_image(image)
        else:
            return image

    def apply_image(self, image: np.ndarray) -> np.ndarray:
        """
        Expects a numpy array with shape HxWxC in uint8 format.
        """
        target_size = self.get_preprocess_shape(image.shape[0], image.shape[1], self.size)
        return np.array(resize(to_pil_image(image), target_size))

    @staticmethod
    def get_preprocess_shape(oldh: int, oldw: int, long_side_length: int) -> tuple[int, int]:
        """
        Compute the output size given input size and target long side length.
        """
        scale = long_side_length * 1.0 / max(oldh, oldw)
        newh, neww = oldh * scale, oldw * scale
        neww = int(neww + 0.5)
        newh = int(newh + 0.5)
        return (newh, neww)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(size={self.size})"


class SamNeck(DAGBlock):
    def __init__(
        self,
        fid_list: list[str],
        in_channel_list: list[int],
        head_width: int,
        head_depth: int,
        expand_ratio: float,
        middle_op: str,
        out_dim: int = 256,
        norm="bn2d",
        act_func="gelu",
    ):
        inputs = {}
        for fid, in_channel in zip(fid_list, in_channel_list):
            inputs[fid] = OpSequential(
                [
                    ConvLayer(in_channel, head_width, 1, norm=norm, act_func=None),
                    UpSampleLayer(size=(64, 64)),
                ]
            )

        middle = []
        for _ in range(head_depth):
            if middle_op == "mb":
                block = MBConv(
                    head_width,
                    head_width,
                    expand_ratio=expand_ratio,
                    norm=norm,
                    act_func=(act_func, act_func, None),
                )
            elif middle_op == "fmb":
                block = FusedMBConv(
                    head_width,
                    head_width,
                    expand_ratio=expand_ratio,
                    norm=norm,
                    act_func=(act_func, None),
                )
            elif middle_op == "res":
                block = ResBlock(
                    head_width,
                    head_width,
                    expand_ratio=expand_ratio,
                    norm=norm,
                    act_func=(act_func, None),
                )
            else:
                raise NotImplementedError
            middle.append(ResidualBlock(block, IdentityLayer()))
        middle = OpSequential(middle)

        outputs = {
            "sam_encoder": OpSequential(
                [
                    ConvLayer(
                        head_width,
                        out_dim,
                        1,
                        use_bias=True,
                        norm=None,
                        act_func=None,
                    ),
                ]
            )
        }

        super(SamNeck, self).__init__(inputs, "add", None, middle=middle, outputs=outputs)


class EfficientViTSamImageEncoder(nn.Module):
    def __init__(self, backbone: EfficientViTBackbone | EfficientViTLargeBackbone, neck: SamNeck):
        super().__init__()
        self.backbone = backbone
        self.neck = neck

        self.norm = build_norm("ln2d", 256)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        feed_dict = self.backbone(x)
        feed_dict = self.neck(feed_dict)

        output = feed_dict["sam_encoder"]
        output = self.norm(output)
        return output


class EfficientViTSam(nn.Module):
    mask_threshold: float = 0.0
    image_format: str = "RGB"

    def __init__(
        self,
        image_encoder: EfficientViTSamImageEncoder,
        prompt_encoder: PromptEncoder,
        mask_decoder: MaskDecoder,
        image_size: tuple[int, int] = (1024, 512),
    ) -> None:
        super().__init__()
        self.image_encoder = image_encoder
        self.prompt_encoder = prompt_encoder
        self.mask_decoder = mask_decoder

        self.image_size = image_size

        self.transform = transforms.Compose(
            [
                SamResize(self.image_size[1]),
                transforms.ToTensor(),
                transforms.Normalize(
                    mean=[123.675 / 255, 116.28 / 255, 103.53 / 255],
                    std=[58.395 / 255, 57.12 / 255, 57.375 / 255],
                ),
                SamPad(self.image_size[1]),
            ]
        )

    def postprocess_masks(
        self,
        masks: torch.Tensor,
        input_size: tuple[int, ...],
        original_size: tuple[int, ...],
    ) -> torch.Tensor:
        masks = F.interpolate(
            masks,
            (self.image_size[0], self.image_size[0]),
            mode="bilinear",
            align_corners=False,
        )
        masks = masks[..., : input_size[0], : input_size[1]]
        masks = F.interpolate(masks, original_size, mode="bilinear", align_corners=False)
        return masks

    def forward(
        self,
        batched_input: list[dict[str, Any]],
        multimask_output: bool,
    ):
        input_images = torch.stack([x["image"] for x in batched_input], dim=0)

        image_embeddings = self.image_encoder(input_images)

        outputs = []
        iou_outputs = []
        for image_record, curr_embedding in zip(batched_input, image_embeddings):
            if "point_coords" in image_record:
                points = (image_record["point_coords"], image_record["point_labels"])
            else:
                points = None
            sparse_embeddings, dense_embeddings = self.prompt_encoder(
                points=points,
                boxes=image_record.get("boxes", None),
                masks=image_record.get("mask_inputs", None),
            )
            low_res_masks, iou_predictions = self.mask_decoder(
                image_embeddings=curr_embedding.unsqueeze(0),
                image_pe=self.prompt_encoder.get_dense_pe(),
                sparse_prompt_embeddings=sparse_embeddings,
                dense_prompt_embeddings=dense_embeddings,
                multimask_output=multimask_output,
            )
            outputs.append(low_res_masks)
            iou_outputs.append(iou_predictions)

        outputs = torch.stack([out for out in outputs], dim=0)
        iou_outputs = torch.stack(iou_outputs, dim=0)

        return outputs, iou_outputs


class EfficientViTSamPredictor:
    def __init__(self, sam_model: EfficientViTSam) -> None:
        self.model = sam_model
        self.reset_image()

    @property
    def transform(self):
        return self

    @property
    def device(self):
        return get_device(self.model)

    def reset_image(self) -> None:
        self.is_image_set = False
        self.features = None
        self.original_size = None
        self.input_size = None

    def apply_coords(self, coords: np.ndarray, im_size=None) -> np.ndarray:
        old_h, old_w = self.original_size
        new_h, new_w = self.input_size
        coords = copy.deepcopy(coords).astype(float)
        coords[..., 0] = coords[..., 0] * (new_w / old_w)
        coords[..., 1] = coords[..., 1] * (new_h / old_h)
        return coords

    def apply_boxes(self, boxes: np.ndarray, im_size=None) -> np.ndarray:
        boxes = self.apply_coords(boxes.reshape(-1, 2, 2))
        return boxes.reshape(-1, 4)

    def apply_boxes_torch(self, boxes: torch.Tensor) -> torch.Tensor:
        boxes = self.apply_coords_torch(boxes.reshape(-1, 2, 2))
        return boxes.reshape(-1, 4)

    def apply_coords_torch(self, coords: torch.Tensor, im_size=None) -> np.ndarray:
        old_h, old_w = self.original_size
        new_h, new_w = self.input_size
        coords_copy = coords.detach().clone().to(torch.float)
        coords_copy[..., 0] = coords_copy[..., 0] * (new_w / old_w)
        coords_copy[..., 1] = coords_copy[..., 1] * (new_h / old_h)
        return coords_copy

    @torch.inference_mode()
    def set_image(self, image: np.ndarray, image_format: str = "RGB") -> None:
        assert image_format in [
            "RGB",
            "BGR",
        ], f"image_format must be in ['RGB', 'BGR'], is {image_format}."
        if image_format != self.model.image_format:
            image = image[..., ::-1]

        self.reset_image()

        self.original_size = image.shape[:2]
        self.input_size = ResizeLongestSide.get_preprocess_shape(
            *self.original_size, long_side_length=self.model.image_size[0]
        )

        torch_data = self.model.transform(image).unsqueeze(dim=0).to(get_device(self.model))
        self.features = self.model.image_encoder(torch_data)
        self.is_image_set = True

    @torch.inference_mode()
    def set_image_batch(self, image: torch.Tensor) -> None:
        """
        image torch.Tensor : Shape (B,C,H,W) with data expected to be preprocessed already, see EfficientViTSam.transform for the expected transforms.
        """
        self.reset_image()

        original_height, original_width = image.shape[-2], image.shape[-1]
        self.original_size = (original_height, original_width)
        self.input_size = ResizeLongestSide.get_preprocess_shape(
            *self.original_size, long_side_length=self.model.image_size[0]
        )

        self.features = self.model.image_encoder(image)
        self.is_image_set = True

    def predict(
        self,
        point_coords: Optional[np.ndarray] = None,
        point_labels: Optional[np.ndarray] = None,
        box: Optional[np.ndarray] = None,
        mask_input: Optional[np.ndarray] = None,
        multimask_output: bool = True,
        return_logits: bool = False,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Predict masks for the given input prompts, using the currently set image.

        Arguments:
          point_coords (np.ndarray or None): A Nx2 array of point prompts to the
            model. Each point is in (X,Y) in pixels.
          point_labels (np.ndarray or None): A length N array of labels for the
            point prompts. 1 indicates a foreground point and 0 indicates a
            background point.
          box (np.ndarray or None): A length 4 array given a box prompt to the
            model, in XYXY format.
          mask_input (np.ndarray): A low resolution mask input to the model, typically
            coming from a previous prediction iteration. Has form 1xHxW, where
            for SAM, H=W=256.
          multimask_output (bool): If true, the model will return three masks.
            For ambiguous input prompts (such as a single click), this will often
            produce better masks than a single prediction. If only a single
            mask is needed, the model's predicted quality score can be used
            to select the best mask. For non-ambiguous prompts, such as multiple
            input prompts, multimask_output=False can give better results.
          return_logits (bool): If true, returns un-thresholded masks logits
            instead of a binary mask.

        Returns:
          (np.ndarray): The output masks in CxHxW format, where C is the
            number of masks, and (H, W) is the original image size.
          (np.ndarray): An array of length C containing the model's
            predictions for the quality of each mask.
          (np.ndarray): An array of shape CxHxW, where C is the number
            of masks and H=W=256. These low resolution logits can be passed to
            a subsequent iteration as mask input.
        """
        if not self.is_image_set:
            raise RuntimeError("An image must be set with .set_image(...) before mask prediction.")

        device = get_device(self.model)
        # Transform input prompts
        coords_torch, labels_torch, box_torch, mask_input_torch = None, None, None, None
        if point_coords is not None:
            assert point_labels is not None, "point_labels must be supplied if point_coords is supplied."
            point_coords = self.apply_coords(point_coords)
            coords_torch = torch.as_tensor(point_coords, dtype=torch.float, device=device)
            labels_torch = torch.as_tensor(point_labels, dtype=torch.int, device=device)
            coords_torch, labels_torch = coords_torch[None, :, :], labels_torch[None, :]
        if box is not None:
            box = self.apply_boxes(box)
            box_torch = torch.as_tensor(box, dtype=torch.float, device=device)
            box_torch = box_torch[None, :]
        if mask_input is not None:
            mask_input_torch = torch.as_tensor(mask_input, dtype=torch.float, device=device)
            mask_input_torch = mask_input_torch[None, :, :, :]

        masks, iou_predictions, low_res_masks = self.predict_torch(
            coords_torch,
            labels_torch,
            box_torch,
            mask_input_torch,
            multimask_output,
            return_logits=return_logits,
        )

        masks = masks[0].detach().cpu().numpy()
        iou_predictions = iou_predictions[0].detach().cpu().numpy()
        low_res_masks = low_res_masks[0].detach().cpu().numpy()
        return masks, iou_predictions, low_res_masks

    @torch.inference_mode()
    def predict_torch(
        self,
        point_coords: Optional[torch.Tensor] = None,
        point_labels: Optional[torch.Tensor] = None,
        boxes: Optional[torch.Tensor] = None,
        mask_input: Optional[torch.Tensor] = None,
        multimask_output: bool = True,
        return_logits: bool = False,
        image_index: Optional[int] = None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Predict masks for the given input prompts, using the currently set image.
        Input prompts are batched torch tensors and are expected to already be
        transformed to the input frame using ResizeLongestSide.
        Arguments:
          point_coords (torch.Tensor or None): A BxNx2 array of point prompts to the
            model. Each point is in (X,Y) in pixels.
          point_labels (torch.Tensor or None): A BxN array of labels for the
            point prompts. 1 indicates a foreground point and 0 indicates a
            background point.
          box (np.ndarray or None): A Bx4 array given a box prompt to the
            model, in XYXY format.
          mask_input (np.ndarray): A low resolution mask input to the model, typically
            coming from a previous prediction iteration. Has form Bx1xHxW, where
            for SAM, H=W=256. Masks returned by a previous iteration of the
            predict method do not need further transformation.
          multimask_output (bool): If true, the model will return three masks.
            For ambiguous input prompts (such as a single click), this will often
            produce better masks than a single prediction. If only a single
            mask is needed, the model's predicted quality score can be used
            to select the best mask. For non-ambiguous prompts, such as multiple
            input prompts, multimask_output=False can give better results.
          return_logits (bool): If true, returns un-thresholded masks logits
            instead of a binary mask.
          image_index (int): If provided will be used to index the image embeddings used by the decoder. This needs to be used if one uses set_image_batch.
        Returns:
          (torch.Tensor): The output masks in BxCxHxW format, where C is the
            number of masks, and (H, W) is the original image size.
          (torch.Tensor): An array of shape BxC containing the model's
            predictions for the quality of each mask.
          (torch.Tensor): An array of shape BxCxHxW, where C is the number
            of masks and H=W=256. These low res logits can be passed to
            a subsequent iteration as mask input.
        """
        if not self.is_image_set:
            raise RuntimeError("An image must be set with .set_image(...) before mask prediction.")
        if point_coords is not None:
            points = (point_coords, point_labels)
        else:
            points = None
        # Embed prompts
        sparse_embeddings, dense_embeddings = self.model.prompt_encoder(
            points=points,
            boxes=boxes,
            masks=mask_input,
        )

        # Predict masks
        if image_index is not None:
            image_embeddings = self.features[image_index].unsqueeze(0)
        else:
            image_embeddings = self.features
        low_res_masks, iou_predictions = self.model.mask_decoder(
            image_embeddings=image_embeddings,
            image_pe=self.model.prompt_encoder.get_dense_pe(),
            sparse_prompt_embeddings=sparse_embeddings,
            dense_prompt_embeddings=dense_embeddings,
            multimask_output=multimask_output,
        )
        # Upscale the masks to the original image resolution
        masks = self.model.postprocess_masks(low_res_masks, self.input_size, self.original_size)
        if not return_logits:
            masks = masks > self.model.mask_threshold
        return masks, iou_predictions, low_res_masks


class EfficientViTSamAutomaticMaskGenerator(SamAutomaticMaskGenerator):
    def __init__(
        self,
        model: EfficientViTSam,
        points_per_side: Optional[int] = 32,
        points_per_batch: int = 64,
        pred_iou_thresh: float = 0.88,
        stability_score_thresh: float = 0.95,
        stability_score_offset: float = 1.0,
        box_nms_thresh: float = 0.7,
        crop_n_layers: int = 0,
        crop_nms_thresh: float = 0.7,
        crop_overlap_ratio: float = 512 / 1500,
        crop_n_points_downscale_factor: int = 1,
        point_grids: Optional[list[np.ndarray]] = None,
        min_mask_region_area: int = 0,
        output_mode: str = "binary_mask",
    ) -> None:
        assert (points_per_side is None) != (
            point_grids is None
        ), "Exactly one of points_per_side or point_grid must be provided."
        if points_per_side is not None:
            self.point_grids = build_all_layer_point_grids(
                points_per_side,
                crop_n_layers,
                crop_n_points_downscale_factor,
            )
        elif point_grids is not None:
            self.point_grids = point_grids
        else:
            raise ValueError("Can't have both points_per_side and point_grid be None.")

        assert output_mode in [
            "binary_mask",
            "uncompressed_rle",
            "coco_rle",
        ], f"Unknown output_mode {output_mode}."

        self.predictor = EfficientViTSamPredictor(model)
        self.points_per_batch = points_per_batch
        self.pred_iou_thresh = pred_iou_thresh
        self.stability_score_thresh = stability_score_thresh
        self.stability_score_offset = stability_score_offset
        self.box_nms_thresh = box_nms_thresh
        self.crop_n_layers = crop_n_layers
        self.crop_nms_thresh = crop_nms_thresh
        self.crop_overlap_ratio = crop_overlap_ratio
        self.crop_n_points_downscale_factor = crop_n_points_downscale_factor
        self.min_mask_region_area = min_mask_region_area
        self.output_mode = output_mode


def build_efficientvit_sam(image_encoder: EfficientViTSamImageEncoder, image_size: int) -> EfficientViTSam:
    return EfficientViTSam(
        image_encoder=image_encoder,
        prompt_encoder=PromptEncoder(
            embed_dim=256,
            image_embedding_size=(64, 64),
            input_image_size=(1024, 1024),
            mask_in_chans=16,
        ),
        mask_decoder=MaskDecoder(
            num_multimask_outputs=3,
            transformer=TwoWayTransformer(
                depth=2,
                embedding_dim=256,
                mlp_dim=2048,
                num_heads=8,
            ),
            transformer_dim=256,
            iou_head_depth=3,
            iou_head_hidden_dim=256,
        ),
        image_size=(1024, image_size),
    )


def efficientvit_sam_l0(image_size: int = 512, **kwargs) -> EfficientViTSam:
    from efficientvit.models.efficientvit.backbone import efficientvit_backbone_l0

    backbone = efficientvit_backbone_l0(**kwargs)

    neck = SamNeck(
        fid_list=["stage4", "stage3", "stage2"],
        in_channel_list=[512, 256, 128],
        head_width=256,
        head_depth=4,
        expand_ratio=1,
        middle_op="fmb",
    )

    image_encoder = EfficientViTSamImageEncoder(backbone, neck)
    return build_efficientvit_sam(image_encoder, image_size)


def efficientvit_sam_l1(image_size: int = 512, **kwargs) -> EfficientViTSam:
    from efficientvit.models.efficientvit.backbone import efficientvit_backbone_l1

    backbone = efficientvit_backbone_l1(**kwargs)

    neck = SamNeck(
        fid_list=["stage4", "stage3", "stage2"],
        in_channel_list=[512, 256, 128],
        head_width=256,
        head_depth=8,
        expand_ratio=1,
        middle_op="fmb",
    )

    image_encoder = EfficientViTSamImageEncoder(backbone, neck)
    return build_efficientvit_sam(image_encoder, image_size)


def efficientvit_sam_l2(image_size: int = 512, **kwargs) -> EfficientViTSam:
    from efficientvit.models.efficientvit.backbone import efficientvit_backbone_l2

    backbone = efficientvit_backbone_l2(**kwargs)

    neck = SamNeck(
        fid_list=["stage4", "stage3", "stage2"],
        in_channel_list=[512, 256, 128],
        head_width=256,
        head_depth=12,
        expand_ratio=1,
        middle_op="fmb",
    )

    image_encoder = EfficientViTSamImageEncoder(backbone, neck)
    return build_efficientvit_sam(image_encoder, image_size)


def efficientvit_sam_xl0(image_size: int = 1024, **kwargs) -> EfficientViTSam:
    from efficientvit.models.efficientvit.backbone import EfficientViTLargeBackbone

    backbone = EfficientViTLargeBackbone(
        width_list=[32, 64, 128, 256, 512, 1024],
        depth_list=[0, 1, 1, 2, 3, 3],
        block_list=["res", "fmb", "fmb", "fmb", "att@3", "att@3"],
        expand_list=[1, 4, 4, 4, 4, 6],
        fewer_norm_list=[False, False, False, False, True, True],
        **build_kwargs_from_config(kwargs, EfficientViTLargeBackbone),
    )

    neck = SamNeck(
        fid_list=["stage5", "stage4", "stage3"],
        in_channel_list=[1024, 512, 256],
        head_width=256,
        head_depth=6,
        expand_ratio=4,
        middle_op="fmb",
    )

    image_encoder = EfficientViTSamImageEncoder(backbone, neck)
    return build_efficientvit_sam(image_encoder, image_size)


def efficientvit_sam_xl1(image_size: int = 1024, **kwargs) -> EfficientViTSam:
    from efficientvit.models.efficientvit.backbone import EfficientViTLargeBackbone

    backbone = EfficientViTLargeBackbone(
        width_list=[32, 64, 128, 256, 512, 1024],
        depth_list=[1, 2, 2, 4, 6, 6],
        block_list=["res", "fmb", "fmb", "fmb", "att@3", "att@3"],
        expand_list=[1, 4, 4, 4, 4, 6],
        fewer_norm_list=[False, False, False, False, True, True],
        **build_kwargs_from_config(kwargs, EfficientViTLargeBackbone),
    )

    neck = SamNeck(
        fid_list=["stage5", "stage4", "stage3"],
        in_channel_list=[1024, 512, 256],
        head_width=256,
        head_depth=12,
        expand_ratio=4,
        middle_op="fmb",
    )

    image_encoder = EfficientViTSamImageEncoder(backbone, neck)
    return build_efficientvit_sam(image_encoder, image_size)



import torch 
from abc import ABC, abstractmethod


class BaseLossWeighting(ABC, torch.nn.Module):
    """
    Defines the abstract class for all the loss weighting classes.
    """
    def __init__(self, objective = "eps"):
        super().__init__()

    @abstractmethod
    def forward(self, timesteps, alphas_cumprod):
        pass

class NoneWeighting(BaseLossWeighting):
    def __init__(self, objective = "eps", *args, **kwargs):
        super().__init__(objective)
        self.objective = objective

    def forward(self, timesteps, alphas_cumprod):
        """
        Calculates the forward pass of the model.

        Args:
            timesteps (torch.Tensor): A tensor of shape (B,) containing the timesteps.
            alphas_cumprod (torch.Tensor): A tensor of shape (T,) containing the cumulative product of alphas.

        Returns:
            torch.Tensor: A tensor of shape (B,) containing the weighted SNR values.
        """
        return torch.ones_like(timesteps)

class SNRWeighting(BaseLossWeighting):
    def __init__(self, objective = "eps", *args, **kwargs):
        """
        Initializes an instance of the SNRWEighting class with the specified objective.

        :param objective: (str) The objective of the class. Defaults to "noise".
        :return: None
        """
        super().__init__(objective)
        self.objective = objective

    def forward(self, timesteps, alphas_cumprod):
        """
        Calculates the forward pass of the model.

        Args:
            timesteps (torch.Tensor): A tensor of shape (B,) containing the timesteps.
            alphas_cumprod (torch.Tensor): A tensor of shape (T,) containing the cumulative product of alphas.

        Returns:
            torch.Tensor: A tensor of shape (B,) containing the weighted SNR values.
        """

        if self.objective == "eps":
            return torch.ones_like(timesteps)
        elif self.objective == "x0":
            sqrt_alphas_cumprod = torch.sqrt(alphas_cumprod)[timesteps]
            sqrt_one_minus_alphas_cumprod = torch.sqrt(1.0 - alphas_cumprod)[timesteps]

            snr = (sqrt_alphas_cumprod / sqrt_one_minus_alphas_cumprod) ** 2
            return snr
        else:
            raise NotImplementedError
        
class InverseSNRWeighting(BaseLossWeighting):
    def __init__(self, objective = "eps", *args, **kwargs):
        """
        Initializes an instance of the SNRWEighting class with the specified objective.

        :param objective: (str) The objective of the class. Defaults to "noise".
        :return: None
        """
        super().__init__(objective)
        self.objective = objective

    def forward(self, timesteps, alphas_cumprod):
        """
        Calculates the forward pass of the model.

        Args:
            timesteps (torch.Tensor): A tensor of shape (B,) containing the timesteps.
            alphas_cumprod (torch.Tensor): A tensor of shape (T,) containing the cumulative product of alphas.

        Returns:
            torch.Tensor: A tensor of shape (B,) containing the weighted SNR values.
        """


        sqrt_alphas_cumprod = torch.sqrt(alphas_cumprod)[timesteps]
        sqrt_one_minus_alphas_cumprod = torch.sqrt(1.0 - alphas_cumprod)[timesteps]

        snr = (sqrt_alphas_cumprod / sqrt_one_minus_alphas_cumprod) ** 2
        return 1 / snr


class EqualWeighting(BaseLossWeighting):
    def __init__(self, objective = "eps", *args, **kwargs):
        """
        Initializes an instance of the SNRWEighting class with the specified objective.

        :param objective: (str) The objective of the class. Defaults to "noise".
        :return: None
        """
        super().__init__(objective)
        self.objective = objective

    def forward(self, timesteps, alphas_cumprod):
        """
        Calculates the forward pass of the model.

        Args:
            timesteps (torch.Tensor): A tensor of shape (B,) containing the timesteps.
            alphas_cumprod (torch.Tensor): A tensor of shape (T,) containing the cumulative product of alphas.

        Returns:
            torch.Tensor: A tensor of shape (B,) containing the weighted SNR values.
        """


        sqrt_alphas_cumprod = torch.sqrt(alphas_cumprod)[timesteps]
        sqrt_one_minus_alphas_cumprod = torch.sqrt(1.0 - alphas_cumprod)[timesteps]

        scale = (sqrt_alphas_cumprod / sqrt_one_minus_alphas_cumprod)
        return 1 / scale
    
class MinSNRWeighting(BaseLossWeighting):
    def __init__(self, objective = "eps", snr_gamma = 5.0):
        """
        Initializes an instance of the MinSNRWeighting class.
        https://arxiv.org/pdf/2303.09556
        https://github.com/TiankaiHang/Min-SNR-Diffusion-Training

        Args:
            objective (str, optional): The objective of the weighting scheme. Defaults to "noise".
            snr_gamma (float, optional): The gamma value for the SNR calculation. Defaults to 5.0.
        """
        super().__init__(objective)
        self.objective = objective
        self.snr_gamma = snr_gamma

    def forward(self, timesteps, alphas_cumprod):
        """
        Calculates the forward pass of the MinSNRWeighting model.

        Args:
            timesteps (torch.Tensor): A tensor of shape (B,) containing the timesteps.
            alphas_cumprod (torch.Tensor): A tensor of shape (T,) containing the cumulative product of alphas.

        Returns:
            torch.Tensor: A tensor of shape (B,) containing the weighted SNR values.

        Raises:
            NotImplementedError: If the objective is not "noise".
        """
        

        sqrt_alphas_cumprod = torch.sqrt(alphas_cumprod)[timesteps]
        sqrt_one_minus_alphas_cumprod = torch.sqrt(1.0 - alphas_cumprod)[timesteps]

        snr = (sqrt_alphas_cumprod / sqrt_one_minus_alphas_cumprod) ** 2

        if self.objective == "eps":
            return torch.stack([snr, self.snr_gamma * torch.ones_like(timesteps)], dim=1).min(dim=1)[0]
        
        elif self.objective == "x0":
            return torch.stack([snr, self.snr_gamma * torch.ones_like(timesteps)], dim=1).min(dim=1)[0] / snr
        else: 
            raise NotImplementedError


import torch 
import torch.nn as nn 
import torch.nn.functional as F

from .modules import MAETokViTEncoder, MAETokViTDecoder
from medlat.modules.alignments import HOGAlignment, DinoAlignment, ClipAlignment
from medlat.registry import register_model
from medlat.utils import init_from_ckpt

__all__ = ['MaskAEModel']

class Normalize(nn.Module):
    def __init__(self, mean, std, device=None):
        super(Normalize, self).__init__()
        if device is None:
            device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.mean = torch.tensor(mean).view(1, -1, 1, 1).to(device)
        self.std = torch.tensor(std).view(1, -1, 1, 1).to(device)

    def forward(self, x):
        return (x - self.mean) / self.std

class Denormalize(nn.Module):
    def __init__(self, mean, std, device=None):
        super(Denormalize, self).__init__()
        if device is None:
            device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.mean = torch.tensor(mean).view(1, -1, 1, 1).to(device)
        self.std = torch.tensor(std).view(1, -1, 1, 1).to(device)

    def forward(self, x):
        return x * self.std + self.mean

def mean_flat(x):
    """
    Take the mean over all non-batch dimensions.
    """
    return torch.mean(x, dim=list(range(1, len(x.size()))))

class MaskAEModel(nn.Module):
    def __init__(self, 
                 img_size: int = 256,
                 patch_size: int = 16,
                 base_img_size: int = 256,
                 codebook_embed_dim: int = 32,
                 dec_use_movq: bool = False,
                 num_latent_tokens: int = 128,
                 to_pixel: str = 'linear',
                 enc_embed_dim: int = 768,
                 dec_embed_dim: int = 768,
                 enc_depth: int = 12,
                 dec_depth: int = 12,
                 enc_num_heads: int = 12,
                 dec_num_heads: int = 12,
                 enc_mlp_ratio: float = 4.,
                 dec_mlp_ratio: float = 4.,
                 enc_drop_path_rate: float = 0.0,
                 dec_drop_path_rate: float = 0.05,
                 dec_cls_token: bool = False,
                 use_ape: bool = False,
                 use_rope: bool = True,
                 rope_mixed: bool = True,
                 rope_theta: float = 10.0,
                 repa_patch_size: int = 16,
                 repa_model: str = 'vit_large_patch14_dinov2.lvd142m',
                 repa_proj_dim: int = 1024,
                 enc_token_drop: float = 0.4,
                 enc_token_drop_max: float = 0.6,
                 aux_dec_model: str = 'vit_tinytiny_patch14_dinov2_movq2',
                 aux_loss_mask: bool = True,
                 aux_dec_cls_token: bool = True,
                 aux_hog_dec: bool = True,
                 aux_dino_dec: bool = True,
                 aux_clip_dec: bool = True,
                 aux_biomed_clip_dec: bool = False,
                 ckpt_path: str = None,
                 ):
        super().__init__()
        self.img_size = img_size
        self.base_img_size = base_img_size
        self.dec_use_movq = dec_use_movq
        self.codebook_embed_dim = codebook_embed_dim
        self.num_latent_tokens = num_latent_tokens
        self.to_pixel = to_pixel
        self.enc_embed_dim = enc_embed_dim
        self.dec_embed_dim = dec_embed_dim
        self.enc_depth = enc_depth
        self.dec_depth = dec_depth
        self.enc_num_heads = enc_num_heads
        self.dec_num_heads = dec_num_heads
        self.enc_mlp_ratio = enc_mlp_ratio
        self.dec_mlp_ratio = dec_mlp_ratio
        self.patch_size = patch_size
        self.enc_drop_path_rate = enc_drop_path_rate
        self.dec_drop_path_rate = dec_drop_path_rate
        self.dec_cls_token = dec_cls_token
        self.use_ape = use_ape
        self.use_rope = use_rope
        self.rope_mixed = rope_mixed
        self.rope_theta = rope_theta
        self.repa_patch_size = repa_patch_size
        self.repa_model = repa_model
        self.repa_proj_dim = repa_proj_dim
        self.enc_token_drop = enc_token_drop
        self.enc_token_drop_max = enc_token_drop_max
        self.aux_dec_model = aux_dec_model  # Keep for backward compatibility but not used directly
        self.aux_loss_mask = aux_loss_mask
        self.aux_dec_cls_token = aux_dec_cls_token
        self.aux_hog_dec = aux_hog_dec
        self.aux_dino_dec = aux_dino_dec
        self.aux_clip_dec = aux_clip_dec
        self.aux_biomed_clip_dec = aux_biomed_clip_dec

        self.encoder = MAETokViTEncoder(
            in_channels=3,
            img_size=self.img_size,
            patch_size=self.patch_size,
            embed_dim=self.enc_embed_dim,
            depth=self.enc_depth,
            num_heads=self.enc_num_heads,
            mlp_ratio=self.enc_mlp_ratio,
            num_latent_tokens=self.num_latent_tokens,
            use_ape=self.use_ape, use_rope=self.use_rope, rope_mixed=self.rope_mixed, rope_theta=self.rope_theta,
            token_drop=self.enc_token_drop,
            token_drop_max=self.enc_token_drop_max,
            base_img_size=self.base_img_size,
            drop_path_rate=self.enc_drop_path_rate
        )

        self.decoder = MAETokViTDecoder(
            in_channels=3,
            img_size=self.img_size,
            patch_size=self.patch_size,
            embed_dim=self.dec_embed_dim,
            depth=self.dec_depth,
            num_heads=self.dec_num_heads,
            mlp_ratio=self.dec_mlp_ratio,
            num_latent_tokens=self.num_latent_tokens,
            use_ape=self.use_ape, use_rope=self.use_rope, rope_mixed=self.rope_mixed, rope_theta=self.rope_theta,
            cls_token=self.dec_cls_token,
            codebook_embed_dim=self.codebook_embed_dim,
            to_pixel=self.to_pixel,
            base_img_size=self.base_img_size,
            use_movq=self.dec_use_movq,
            drop_path_rate=self.dec_drop_path_rate
        )           
        # Add missing quant_conv layer
        self.quant_conv = nn.Linear(self.encoder.embed_dim, self.codebook_embed_dim)
        
        # Add missing post_quant_conv for main decoder
        self.post_quant_conv = nn.Linear(self.codebook_embed_dim, self.decoder.embed_dim)
        
        # Initialize alignment modules
        self.aux_hog_decoder = None
        if self.aux_hog_dec:
            print('Using HOG decoder')
            use_movq_hog = 'movq' in self.aux_dec_model if self.aux_dec_model else False
            # Create decoder for HOG alignment
            aux_hog_decoder_model = MAETokViTDecoder(
                in_channels=3,
                embed_dim=384,  # Default decoder dimensions
                depth=12,
                num_heads=6,
                mlp_ratio=4.0,
                img_size=self.img_size,
                patch_size=self.patch_size,
                drop_path_rate=0.0,
                num_latent_tokens=self.num_latent_tokens,
                to_pixel='identity',
                codebook_embed_dim=self.codebook_embed_dim,
                rope_theta=self.rope_theta,
                rope_mixed=self.rope_mixed,
                use_rope=self.use_rope,
                use_ape=self.use_ape,
                cls_token=self.aux_dec_cls_token,
                base_img_size=self.base_img_size,
                use_movq=use_movq_hog,
            )
            self.aux_hog_decoder = HOGAlignment(
                decoder=aux_hog_decoder_model,
                codebook_embed_dim=self.codebook_embed_dim,
                use_movq=use_movq_hog,
            )
        
        self.aux_dino_decoder = None
        if self.aux_dino_dec:
            print('Using DINO decoder')
            use_movq_dino = 'movq' in self.aux_dec_model if self.aux_dec_model else False
            # Create decoder for DINO alignment
            aux_dino_decoder_model = MAETokViTDecoder(
                in_channels=3,
                embed_dim=384,
                depth=12,
                num_heads=6,
                mlp_ratio=4.0,
                img_size=self.img_size,
                patch_size=self.repa_patch_size,
                drop_path_rate=0.0,
                num_latent_tokens=self.num_latent_tokens,
                to_pixel='identity',
                codebook_embed_dim=self.codebook_embed_dim,
                rope_theta=self.rope_theta,
                rope_mixed=self.rope_mixed,
                use_rope=self.use_rope,
                use_ape=self.use_ape,
                cls_token=self.aux_dec_cls_token,
                base_img_size=self.base_img_size,
                use_movq=use_movq_dino,
            )
            self.aux_dino_decoder = DinoAlignment(
                decoder=aux_dino_decoder_model,
                codebook_embed_dim=self.codebook_embed_dim,
                img_size=self.img_size,
                repa_model_name=self.repa_model,
                repa_patch_size=self.repa_patch_size,
                use_movq=use_movq_dino,
            )
        
        self.aux_clip_decoder = None
        if self.aux_clip_dec:
            print('Using CLIP decoder')
            use_movq_clip = 'movq' in self.aux_dec_model if self.aux_dec_model else False
            # Create decoder for CLIP alignment
            aux_clip_decoder_model = MAETokViTDecoder(
                in_channels=3,
                embed_dim=384,
                depth=12,
                num_heads=6,
                mlp_ratio=4.0,
                img_size=self.img_size,
                patch_size=self.repa_patch_size,
                drop_path_rate=0.0,
                num_latent_tokens=self.num_latent_tokens,
                to_pixel='identity',
                codebook_embed_dim=self.codebook_embed_dim,
                rope_theta=self.rope_theta,
                rope_mixed=self.rope_mixed,
                use_rope=self.use_rope,
                use_ape=self.use_ape,
                cls_token=self.aux_dec_cls_token,
                base_img_size=self.base_img_size,
                use_movq=use_movq_clip,
            )
            self.aux_clip_decoder = ClipAlignment(
                decoder=aux_clip_decoder_model,
                codebook_embed_dim=self.codebook_embed_dim,
                img_size=self.img_size,
                clip_model_name='vit_so400m_patch14_siglip_gap_224',
                clip_patch_size=self.repa_patch_size,
                use_movq=use_movq_clip,
            )
            
        # self.aux_biomed_clip_decoder = self.aux_biomed_clip_dec
        # if self.aux_biomed_clip_decoder:
        #     self.biomed_clip_model, _, _ = open_clip.create_model_and_transforms(
        #     model_name="hf-hub:microsoft/BiomedCLIP-PubMedBERT_256-vit_base_patch16_224",
        # )
        #     for param in self.biomed_clip_model.parameters():
        #         param.requires_grad = False
        #     self.biomed_clip_model.dynamic_img_size = True
        #     self.biomed_clip_model.eval()

        #     print('Using Biomed CLIP decoder: ', self.aux_dec_model)
        #     self.decoder_biomed_clip = MAETokViTDecoder(
        #         in_channels=3,
        #         img_size=[224,352],
        #         patch_size=repa_patch_size,
        #         embed_dim=384,
        #         depth=6,
        #         num_heads=6,
        #         mlp_ratio=4.,
        #         num_latent_tokens=self.num_latent_tokens,
        #         use_ape=self.use_ape, use_rope=self.use_rope, rope_mixed=self.rope_mixed, rope_theta=self.rope_theta,
        #         cls_token=self.aux_dec_cls_token,
        #         codebook_embed_dim=self.codebook_embed_dim,
        #         to_pixel='identity',
        #         base_img_size=self.base_img_size,
        #         use_movq=self.dec_use_movq,
        #     )           
        #     self.post_quant_conv_clip = nn.Linear(self.codebook_embed_dim, self.decoder_biomed_clip.embed_dim)
        #     self.to_pixel_clip = nn.Linear(self.decoder_biomed_clip.embed_dim, self.biomed_clip_model.embed_dim)

        if ckpt_path is not None:
            init_from_ckpt(self, ckpt_path)


    def encode(self, x):
        if self.training:
            h, mask = self.encoder(x, return_mask=True)
        else:
            h = self.encoder(x)
        quant = self.quant_conv(h)
        emb_loss = torch.tensor(0.)
        info = None
        if self.training:
            return quant, emb_loss, info, mask
        else:
            return quant, emb_loss, info

    def decode(self, quant, x=None, h=None, w=None):
        tmp_quant = quant 
        quant = self.post_quant_conv(quant)
        if self.dec_use_movq:
            dec = self.decoder(quant, interpolate_zq=tmp_quant, H=h, W=w, D=None)
        else:
            dec = self.decoder(quant, interpolate_zq=None, H=h, W=w, D=None)
        return dec

    def forward(self, input):
        b, _, h, w, *_  = input.size()
        if self.training:
            quant, diff, info, mask = self.encode(input)
        else:
            quant, diff, info = self.encode(input)
        self.quant = quant
        dec = self.decode(quant, x=input)
        # Compute alignment losses using alignment modules
        if self.training:
            ## get rid of the cls token in mask 
            if mask is not None:
                mask = mask[:, 1:, :]
            # HOG alignment loss
            if self.aux_hog_decoder is not None:
                hog_loss, _ = self.aux_hog_decoder(quant, input_image=input, mask=mask if self.aux_loss_mask else None)
                hog_rec_loss = hog_loss
            else:
                hog_rec_loss = torch.tensor(0.0, device=input.device)
            
            # DINO alignment loss
            if self.aux_dino_decoder is not None:
                dino_loss, _ = self.aux_dino_decoder(quant, input_image=input, mask=mask if self.aux_loss_mask else None)
                dino_rec_loss = dino_loss
            else:
                dino_rec_loss = torch.tensor(0.0, device=input.device)
            
            # CLIP alignment loss
            if self.aux_clip_decoder is not None:
                clip_loss, _ = self.aux_clip_decoder(quant, input_image=input, mask=mask if self.aux_loss_mask else None)
                clip_rec_loss = clip_loss
            else:
                clip_rec_loss = torch.tensor(0.0, device=input.device)
        
        # Combine auxiliary losses with main loss
        if self.training:
            total_aux_loss = hog_rec_loss + dino_rec_loss + clip_rec_loss
            diff = diff + total_aux_loss
        
        return dec, diff

    def get_last_layer(self):
        # Access the last layer of the to_pixel projection
        if hasattr(self.decoder.to_pixel, 'proj'):
            if isinstance(self.decoder.to_pixel.proj, nn.Linear):
                return self.decoder.to_pixel.proj
            elif isinstance(self.decoder.to_pixel.proj, (nn.Conv2d, nn.Conv3d)):
                return self.decoder.to_pixel.proj
        return None



import collections
import os
from inspect import signature
from typing import Any, Callable, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = [
    "is_parallel",
    "get_device",
    "get_same_padding",
    "resize",
    "build_kwargs_from_config",
    "load_state_dict_from_file",
    "get_submodule_weights",
]


def is_parallel(model: nn.Module) -> bool:
    return isinstance(model, (nn.parallel.DataParallel, nn.parallel.DistributedDataParallel))


def get_device(model: nn.Module) -> torch.device:
    return model.parameters().__next__().device


def get_dtype(model: nn.Module) -> torch.dtype:
    return model.parameters().__next__().dtype


def get_same_padding(kernel_size: int | tuple[int, ...]) -> int | tuple[int, ...]:
    if isinstance(kernel_size, tuple):
        return tuple([get_same_padding(ks) for ks in kernel_size])
    else:
        assert kernel_size % 2 > 0, "kernel size should be odd number"
        return kernel_size // 2


def resize(
    x: torch.Tensor,
    size: Optional[Any] = None,
    scale_factor: Optional[list[float]] = None,
    mode: str = "bicubic",
    align_corners: Optional[bool] = False,
) -> torch.Tensor:
    if mode in {"bilinear", "bicubic"}:
        return F.interpolate(
            x,
            size=size,
            scale_factor=scale_factor,
            mode=mode,
            align_corners=align_corners,
        )
    elif mode in {"nearest", "area"}:
        return F.interpolate(x, size=size, scale_factor=scale_factor, mode=mode)
    else:
        raise NotImplementedError(f"resize(mode={mode}) not implemented.")


def build_kwargs_from_config(config: dict, target_func: Callable) -> dict[str, Any]:
    valid_keys = list(signature(target_func).parameters)
    kwargs = {}
    for key in config:
        if key in valid_keys:
            kwargs[key] = config[key]
    return kwargs


def load_state_dict_from_file(file: str, only_state_dict=True) -> dict[str, torch.Tensor]:
    file = os.path.realpath(os.path.expanduser(file))
    checkpoint = torch.load(file, map_location="cpu", weights_only=True)
    if only_state_dict and "state_dict" in checkpoint:
        checkpoint = checkpoint["state_dict"]
    return checkpoint


def get_submodule_weights(weights: collections.OrderedDict, prefix: str):
    submodule_weights = collections.OrderedDict()
    len_prefix = len(prefix)
    for key, weight in weights.items():
        if key.startswith(prefix):
            submodule_weights[key[len_prefix:]] = weight
    return submodule_weights


def get_dtype_from_str(dtype: str) -> torch.dtype:
    if dtype == "fp32":
        return torch.float32
    if dtype == "fp16":
        return torch.float16
    if dtype == "bf16":
        return torch.bfloat16
    raise NotImplementedError(f"dtype {dtype} is not supported")



from .openaimodel import UNetModel
from medlat.registry import register_model

"""
Model configurations as taken from Table 14 in the paper
"""

@register_model("ldm.f1")
def LDM_f1(img_size, in_channels=3, out_channels=3, model_channels=192, num_head_channels=32, **kwargs):
    return UNetModel(
        img_size=img_size,
        in_channels=in_channels,
        model_channels=model_channels,
        out_channels=out_channels,
        num_res_blocks=2,
        attention_resolutions=[32, 16, 8],
        channel_mult=[1,1,2,2,4,4],
        num_head_channels=num_head_channels,
        **kwargs
    )

@register_model("ldm.f2")
def LDM_f2(img_size, in_channels=3, out_channels=3, model_channels=192, num_head_channels=32, **kwargs):
    return UNetModel(
        img_size=img_size,
        in_channels=in_channels,
        model_channels=model_channels,
        out_channels=out_channels,
        num_res_blocks=2,
        attention_resolutions=[32, 16, 8],
        channel_mult=[1,2,2,4],
        num_head_channels=num_head_channels,
        **kwargs
    )

@register_model("ldm.f4")
def LDM_f4(img_size, in_channels=3, out_channels=3, model_channels=224, num_head_channels=32, **kwargs):
    return UNetModel(
        img_size=img_size,
        in_channels=in_channels,
        model_channels=model_channels,
        out_channels=out_channels,
        num_res_blocks=2,
        attention_resolutions=[32, 16, 8],
        channel_mult=[1,2,3,4],
        num_head_channels=num_head_channels,
        **kwargs
    )

@register_model("ldm.f8")
def LDM_f8(img_size, in_channels=4, out_channels=4, model_channels=256, num_head_channels=32, **kwargs):
    return UNetModel(
        img_size=img_size,
        in_channels=in_channels,
        model_channels=model_channels,
        out_channels=out_channels,
        num_res_blocks=2,
        attention_resolutions=[32, 16, 8],
        channel_mult=[1,2,4],
        num_head_channels=num_head_channels,
        **kwargs
    )

@register_model("ldm.f16")
def LDM_f16(img_size, in_channels=4, out_channels=4, model_channels=256, num_head_channels=32, **kwargs):
    return UNetModel(
        img_size=img_size,
        in_channels=in_channels,
        model_channels=model_channels,
        out_channels=out_channels,
        num_res_blocks=2,
        attention_resolutions=[16, 8, 4],
        channel_mult=[1,2,4],
        num_head_channels=num_head_channels,
        **kwargs
    )





import torch
import torch.nn as nn

from .backbone import EfficientViTBackbone, EfficientViTLargeBackbone
from ..nn import ConvLayer, LinearLayer, OpSequential
from ..utils.network import build_kwargs_from_config

__all__ = [
    "EfficientViTCls",
    ######################
    "efficientvit_cls_b0",
    "efficientvit_cls_b1",
    "efficientvit_cls_b2",
    "efficientvit_cls_b3",
    ######################
    "efficientvit_cls_l1",
    "efficientvit_cls_l2",
    "efficientvit_cls_l3",
]


class ClsHead(OpSequential):
    def __init__(
        self,
        in_channels: int,
        width_list: list[int],
        n_classes=1000,
        dropout=0.0,
        norm="bn2d",
        act_func="hswish",
        fid="stage_final",
    ):
        ops = [
            ConvLayer(in_channels, width_list[0], 1, norm=norm, act_func=act_func),
            nn.AdaptiveAvgPool2d(output_size=1),
            LinearLayer(width_list[0], width_list[1], False, norm="ln", act_func=act_func),
            LinearLayer(width_list[1], n_classes, True, dropout, None, None),
        ]
        super().__init__(ops)

        self.fid = fid

    def forward(self, feed_dict: dict[str, torch.Tensor]) -> torch.Tensor:
        x = feed_dict[self.fid]
        return OpSequential.forward(self, x)


class EfficientViTCls(nn.Module):
    def __init__(self, backbone: EfficientViTBackbone | EfficientViTLargeBackbone, head: ClsHead) -> None:
        super().__init__()
        self.backbone = backbone
        self.head = head

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        feed_dict = self.backbone(x)
        output = self.head(feed_dict)
        return output


def efficientvit_cls_b0(**kwargs) -> EfficientViTCls:
    from efficientvit.models.efficientvit.backbone import efficientvit_backbone_b0

    backbone = efficientvit_backbone_b0(**kwargs)

    head = ClsHead(
        in_channels=128,
        width_list=[1024, 1280],
        **build_kwargs_from_config(kwargs, ClsHead),
    )
    model = EfficientViTCls(backbone, head)
    return model


def efficientvit_cls_b1(**kwargs) -> EfficientViTCls:
    from .backbone import efficientvit_backbone_b1

    backbone = efficientvit_backbone_b1(**kwargs)

    head = ClsHead(
        in_channels=256,
        width_list=[1536, 1600],
        **build_kwargs_from_config(kwargs, ClsHead),
    )
    model = EfficientViTCls(backbone, head)
    return model


def efficientvit_cls_b2(**kwargs) -> EfficientViTCls:
    from efficientvit.models.efficientvit.backbone import efficientvit_backbone_b2

    backbone = efficientvit_backbone_b2(**kwargs)

    head = ClsHead(
        in_channels=384,
        width_list=[2304, 2560],
        **build_kwargs_from_config(kwargs, ClsHead),
    )
    model = EfficientViTCls(backbone, head)
    return model


def efficientvit_cls_b3(**kwargs) -> EfficientViTCls:
    from .backbone import efficientvit_backbone_b3

    backbone = efficientvit_backbone_b3(**kwargs)

    head = ClsHead(
        in_channels=512,
        width_list=[2304, 2560],
        **build_kwargs_from_config(kwargs, ClsHead),
    )
    model = EfficientViTCls(backbone, head)
    return model


def efficientvit_cls_l1(**kwargs) -> EfficientViTCls:
    from .backbone import efficientvit_backbone_l1

    backbone = efficientvit_backbone_l1(**kwargs)

    head = ClsHead(
        in_channels=512,
        width_list=[3072, 3200],
        act_func="gelu",
        **build_kwargs_from_config(kwargs, ClsHead),
    )
    model = EfficientViTCls(backbone, head)
    return model


def efficientvit_cls_l2(**kwargs) -> EfficientViTCls:
    from efficientvit.models.efficientvit.backbone import efficientvit_backbone_l2

    backbone = efficientvit_backbone_l2(**kwargs)

    head = ClsHead(
        in_channels=512,
        width_list=[3072, 3200],
        act_func="gelu",
        **build_kwargs_from_config(kwargs, ClsHead),
    )
    model = EfficientViTCls(backbone, head)
    return model


def efficientvit_cls_l3(**kwargs) -> EfficientViTCls:
    from .backbone import efficientvit_backbone_l3

    backbone = efficientvit_backbone_l3(**kwargs)

    head = ClsHead(
        in_channels=1024,
        width_list=[6144, 6400],
        act_func="gelu",
        **build_kwargs_from_config(kwargs, ClsHead),
    )
    model = EfficientViTCls(backbone, head)
    return model



from typing import Any, Optional

__all__ = [
    "list_sum",
    "list_mean",
    "weighted_list_sum",
    "list_join",
    "val2list",
    "val2tuple",
    "squeeze_list",
]


def list_sum(x: list) -> Any:
    return x[0] if len(x) == 1 else x[0] + list_sum(x[1:])


def list_mean(x: list) -> Any:
    return list_sum(x) / len(x)


def weighted_list_sum(x: list, weights: list) -> Any:
    assert len(x) == len(weights)
    return x[0] * weights[0] if len(x) == 1 else x[0] * weights[0] + weighted_list_sum(x[1:], weights[1:])


def list_join(x: list, sep="\t", format_str="%s") -> str:
    return sep.join([format_str % val for val in x])


def val2list(x: list | tuple | Any, repeat_time=1) -> list:
    if isinstance(x, (list, tuple)):
        return list(x)
    return [x for _ in range(repeat_time)]


def val2tuple(x: list | tuple | Any, min_len: int = 1, idx_repeat: int = -1) -> tuple:
    x = val2list(x)

    # repeat elements if necessary
    if len(x) > 0:
        x[idx_repeat:idx_repeat] = [x[idx_repeat] for _ in range(min_len - len(x))]

    return tuple(x)


def squeeze_list(x: Optional[list]) -> list | Any:
    if x is not None and len(x) == 1:
        return x[0]
    else:
        return x



"""This file contains the model definition of TiTok.

Copyright (2024) Bytedance Ltd. and/or its affiliates

Licensed under the Apache License, Version 2.0 (the "License"); 
you may not use this file except in compliance with the License. 
You may obtain a copy of the License at 

    http://www.apache.org/licenses/LICENSE-2.0 

Unless required by applicable law or agreed to in writing, software 
distributed under the License is distributed on an "AS IS" BASIS, 
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
See the License for the specific language governing permissions and 
limitations under the License.
"""

""" 
taken from http://github.com/bytedance/1d-tokenizer/blob/main/modeling/titok.py

"""

import torch
import torch.nn as nn
from einops import rearrange
import torch.nn.functional as F
from medlat.first_stage.token.titok.modules import TiTokEncoder, TiTokDecoder, Pixel_Encoder, Pixel_Decoder, Pixel_Quantizer
from medlat.first_stage.modules.gaussian_dist import DiagonalGaussianDistribution
import json
from omegaconf import OmegaConf
from pathlib import Path
from medlat.first_stage.discrete.vq_models import VQModel
from medlat.first_stage.discrete.quantizer.quantize import VectorQuantizer2
from typing import Optional
from medlat.utils import init_from_ckpt
__all__ = ["TiTok"]




class PretrainedTokenizer(nn.Module):
    def __init__(self, pretrained_weight):
        super().__init__()
        conf = OmegaConf.create(
            {"channel_mult": [1, 1, 2, 2, 4],
            "num_resolutions": 5,
            "dropout": 0.0,
            "hidden_channels": 128,
            "num_channels": 3,
            "num_res_blocks": 2,
            "resolution": 256,
            "z_channels": 256})
        self.encoder = Pixel_Encoder(conf)
        self.decoder = Pixel_Decoder(conf)
        self.quantize = Pixel_Quantizer(
            num_embeddings=1024, embedding_dim=256, commitment_cost=0.25)
        # Load pretrained weights
        self.load_state_dict(torch.load(pretrained_weight, map_location=torch.device("cpu"), weights_only=False), strict=True)
        print("Loaded pretrained weights")
        
        self.eval()
        for param in self.parameters():
            param.requires_grad = False
    
    @torch.no_grad()
    def encode(self, x):
        hidden_states = self.encoder(x)
        quantized_states, codebook_loss, (_,_, codebook_indices) = self.quantize(hidden_states)
        return quantized_states, codebook_loss, (None,None, codebook_indices)
    
    @torch.no_grad()
    def decode_code(self, codebook_indices, **kwargs):
        return self.decode(codebook_indices)
    

    @torch.no_grad()
    def decode(self, codes):
        quantized_states = self.quantize.get_codebook_entry(codes)
        rec_images = self.decoder(quantized_states)
        rec_images = torch.clamp(rec_images, 0.0, 1.0)
        return rec_images.detach()
    
    @torch.no_grad()
    def decode_tokens(self, codes):
        return self.decode(codes)



"""
Important NOTE: 
The token size should be thought about together with the quantizer object!

Output of the encoder: [B, C=token_size, 1 = num_latent_tokens]
In the quantizer the shape is rearranged to [B, 1, num_latent_tokens, token_size]
In the quantizer the shape is flattend to [-1, e_dim] 
implicitly translating to two cases: 

You can think of TiTOK as a learned, low-dimensional interface into a richer codebook:
- The codebook has many expressive atoms (e.g. 1024 entries of 256 dimensions)
- The encoder finds just a few best-fitting atoms to describe the whole image
- The decoder (or transformer) then decodes the image from those sparse tokens

"""


class TiTok(nn.Module):
    def __init__(self, img_size=256,
                 patch_size=16, 
                 hidden_size=768,
                 in_channels=3,
                 out_channels=3,
                 depth=12,
                 num_heads=12,
                 num_latent_tokens=64,
                 token_size=12,
                 codebook_size=4096,
                 quantizer_loss_weight=1.0,
                 pixel_vqgan: Optional[VQModel] = None, # VQModel = PretrainedTokenizer("/vol/miltank/users/bubeckn/1d-tokenizer/maskgit-vqgan-imagenet-f16-256.bin"), 
                 stage="1", 
                 quantize_mode="vq",
                 dims=2,
                 ckpt_path=None):
        super().__init__()
        # This should be False for stage1 and True for stage2.
        self.stage=stage
        assert stage in ["1", "2", "e2e"]
        if isinstance(img_size, int):
            img_size = (img_size,) * dims
        if isinstance(patch_size, int):
            patch_size = (patch_size,) * dims

        assert len(img_size) == len(patch_size)
        assert not (self.stage == "1" and pixel_vqgan is None), "For stage 1, pixel_vqgan is required."
        self.dims = len(img_size)

        self.n_embed = codebook_size
        self.embed_dim = token_size
        print(f"num latent tokens: {num_latent_tokens}")
        self.quantize_mode = quantize_mode
        self.quantizer_loss_weight = quantizer_loss_weight
        if self.quantize_mode not in ["vq", "vae"]:
            raise ValueError(f"Unsupported quantize mode {self.quantize_mode}.")
        if self.stage == "2" and self.quantize_mode not in ["vq"]:
            raise ValueError("Only support finetune_decoder with vq quantization for now.")
                
        self.encoder = TiTokEncoder(
            img_size=img_size,
            in_channels=in_channels,
            patch_size=patch_size,
            hidden_size=hidden_size,
            depth=depth,
            num_heads=num_heads,
            num_latent_tokens=num_latent_tokens,
            token_size=token_size,
            quantize_mode=quantize_mode,
            is_legacy= not self.stage == "e2e",
        )
        cs = pixel_vqgan.quantize.n_e if pixel_vqgan is not None and self.stage != "e2e" else codebook_size
        self.decoder = TiTokDecoder(
            img_size=img_size,
            out_channels=out_channels,
            codebook_size=cs,
            patch_size=patch_size,
            hidden_size=hidden_size,
            depth=depth,
            num_heads=num_heads,
            num_latent_tokens=num_latent_tokens,
            token_size=token_size,
            quantize_mode=quantize_mode,
            is_legacy= not self.stage == "e2e",
        )
        
        self.num_latent_tokens = num_latent_tokens
        scale = self.encoder.hidden_size ** -0.5
        self.latent_tokens = nn.Parameter(
            scale * torch.randn(self.num_latent_tokens, self.encoder.hidden_size))
        self.apply(self._init_weights)

        if self.quantize_mode == "vq":
            self.quantize = VectorQuantizer2(
                n_e=codebook_size,
                e_dim=token_size,
                legacy=False, # fixes beta weighting error of taming
                beta=0.25,
                use_norm=True,
                )
        elif self.quantize_mode == "vae":
            self.quantize = DiagonalGaussianDistribution
        else:
            raise NotImplementedError
        
        if self.stage == "2":
            # Freeze encoder/quantizer/latent tokens
            self.latent_tokens.requires_grad_(False)
            self.encoder.eval()
            self.encoder.requires_grad_(False)
            self.quantize.eval()
            self.quantize.requires_grad_(False)

            self.pixel_quantize = Pixel_Quantizer(
                num_embeddings=1024, embedding_dim=256, commitment_cost=0.25)
            self.pixel_decoder = Pixel_Decoder(OmegaConf.create(
                {"channel_mult": [1, 1, 2, 2, 4],
                "num_resolutions": 5,
                "dropout": 0.0,
                "hidden_channels": 128,
                "num_channels": 3,
                "num_res_blocks": 2,
                "resolution": 256,
                "z_channels": 256}))
            
        # Freeze the pixel_vqgan is frozen when loaded
        if pixel_vqgan is not None:
            self.pixel_vqgan = pixel_vqgan.eval()
            self.pixel_vqgan.requires_grad_(False)

        if ckpt_path is not None:
            init_from_ckpt(self, ckpt_path)
        
    def _save_pretrained(self, save_directory: Path) -> None:
        """Save weights and config to a local directory."""
        # Assume 'self.config' is your DictConfig object
        # Convert to a regular dictionary
        dict_config = OmegaConf.to_container(self.config)
        # Save as JSON
        file_path = Path(save_directory) / "config.json"
        with open(file_path, 'w') as json_file:
            json.dump(dict_config, json_file, indent=4)
        super()._save_pretrained(save_directory)

    def _init_weights(self, module):
        """ Initialize the weights.
            :param:
                module -> torch.nn.Module: module to initialize
        """
        if isinstance(module, nn.Linear) or isinstance(module, nn.Conv1d) or isinstance(module, nn.Conv2d):
            module.weight.data = nn.init.trunc_normal_(module.weight.data, mean=0.0, std=0.02)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data = nn.init.trunc_normal_(module.weight.data, mean=0.0, std=0.02)
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def encode(self, x):
        if self.stage == "2":
            with torch.no_grad():
                self.encoder.eval()
                self.quantize.eval()
                z = self.encoder(pixel_values=x, latent_tokens=self.latent_tokens)
                z_quantized, loss, (perplexity, min_encodings, min_encoding_indices) = self.quantize(z)
                loss = torch.tensor(0.0)
        else:
            z = self.encoder(pixel_values=x, latent_tokens=self.latent_tokens)
            if self.quantize_mode == "vq":
                z_quantized, loss, (perplexity, min_encodings, min_encoding_indices) = self.quantize(z)
            elif self.quantize_mode == "vae":
                posteriors = self.quantize(z)
                z_quantized = posteriors.sample()
                # result_dict = posteriors
        return z_quantized, loss, (perplexity, min_encodings, min_encoding_indices)
    
    
    def decode(self, z_quantized): 
        decoded = self.decoder(z_quantized)
        if self.stage == "2":
            quantized_states = torch.einsum(
                'nchw,cd->ndhw', decoded.softmax(1),
                self.pixel_quantize.embedding.weight)
            decoded = self.pixel_decoder(quantized_states)
        return decoded
    

    def decode_code(self, tokens, out_shape=None):
        if self.quantize_mode == "vq":
            tokens = tokens.squeeze(1)
            quant_b = self.quantize.get_codebook_entry(tokens, shape=out_shape)
            # Move channel dimension (which is last) to the second
            if quant_b.dim() == 4:
                # (B, H, W, C) -> (B, C, H, W)
                quant_b = quant_b.permute(0, 3, 1, 2).contiguous()
            elif quant_b.dim() == 5:
                # (B, D, H, W, C) -> (B, C, D, H, W)
                quant_b = quant_b.permute(0, 4, 1, 2, 3).contiguous()
        elif self.quantize_mode == "vae":
            quant_b = tokens
        decoded = self.decode(quant_b)
        return decoded
    
    def proxy_code_loss(self, decoded, x, q_loss):
        _, _, (_, _, idx) = self.pixel_vqgan.encode(x)
        idx = idx.reshape(decoded.shape[0], -1) # B x N
        decoded = decoded.reshape(decoded.shape[0], decoded.shape[1], -1)
        print(idx.shape, decoded.shape)
        ce_loss = F.cross_entropy(decoded, idx, reduction='mean')
        total_loss = ce_loss + q_loss * self.quantizer_loss_weight
        return total_loss.float()
    
    
    def forward(self, x):
        z_quantized, loss, (_, _, _) = self.encode(x)
        decoded = self.decode(z_quantized)
        return decoded, loss

    


from functools import partial

import numpy as np
from tqdm import tqdm
import scipy.stats as stats
import math
import torch
import torch.nn as nn
import random
from torch.utils.checkpoint import checkpoint

from timm.models.vision_transformer import Block

from medlat.diffusion.diffloss import DiffLoss

__all__ = ['MAR']


def mask_by_order(mask_len, order, bsz, seq_len):
    masking = torch.zeros(bsz, seq_len).cuda()
    masking = torch.scatter(masking, dim=-1, index=order[:, :mask_len.long()], src=torch.ones(bsz, seq_len).cuda()).bool()
    return masking


class MAR(nn.Module):
    """ Masked Autoencoder with VisionTransformer backbone
    """
    def __init__(self, img_size=256, vae_stride=16, patch_size=1,
                 encoder_embed_dim=1024, encoder_depth=16, encoder_num_heads=16,
                 decoder_embed_dim=1024, decoder_depth=16, decoder_num_heads=16,
                 mlp_ratio=4., norm_layer=nn.LayerNorm,
                 in_channels=16,
                 mask_ratio_min=0.7,
                 label_drop_prob=0.1,
                 class_num=1000,
                 dataset_num=None,
                 attn_dropout=0.1,
                 proj_dropout=0.1,
                 buffer_size=64,
                 diffloss_d=3,
                 diffloss_w=1024,
                 num_sampling_steps='100',
                 diffusion_batch_mul=4,
                 grad_checkpointing=False,
                 ):
        super().__init__()

        # --------------------------------------------------------------------------
        # VAE and patchify specifics
        self.in_channels = in_channels

        self.img_size = img_size
        self.vae_stride = vae_stride
        self.patch_size = patch_size
        self.seq_h = self.seq_w = img_size // vae_stride // patch_size
        self.seq_len = self.seq_h * self.seq_w
        self.token_embed_dim = in_channels * patch_size**2
        self.grad_checkpointing = grad_checkpointing

        # --------------------------------------------------------------------------
        # Class Embedding
        self.num_classes = class_num
        self.class_emb = nn.Embedding(class_num, encoder_embed_dim)
        self.label_drop_prob = label_drop_prob
        # Fake class embedding for CFG's unconditional generation
        self.fake_latent = nn.Parameter(torch.zeros(1, encoder_embed_dim))
        
        # --------------------------------------------------------------------------
        # Dataset ID Embedding (optional)
        self.use_dataset_conditioning = dataset_num is not None
        if self.use_dataset_conditioning:
            self.num_datasets = dataset_num
            self.dataset_emb = nn.Embedding(dataset_num, encoder_embed_dim)
            # Fake dataset embedding for unconditional generation
            self.fake_dataset_latent = nn.Parameter(torch.zeros(1, encoder_embed_dim))

        # --------------------------------------------------------------------------
        # MAR variant masking ratio, a left-half truncated Gaussian centered at 100% masking ratio with std 0.25
        self.mask_ratio_generator = stats.truncnorm((mask_ratio_min - 1.0) / 0.25, 0, loc=1.0, scale=0.25)

        # --------------------------------------------------------------------------
        # MAR encoder specifics
        self.z_proj = nn.Linear(self.token_embed_dim, encoder_embed_dim, bias=True)
        self.z_proj_ln = nn.LayerNorm(encoder_embed_dim, eps=1e-6)
        self.buffer_size = buffer_size
        self.encoder_pos_embed_learned = nn.Parameter(torch.zeros(1, self.seq_len + self.buffer_size, encoder_embed_dim))

        self.encoder_blocks = nn.ModuleList([
            Block(encoder_embed_dim, encoder_num_heads, mlp_ratio, qkv_bias=True, norm_layer=norm_layer,
                  proj_drop=proj_dropout, attn_drop=attn_dropout) for _ in range(encoder_depth)])
        self.encoder_norm = norm_layer(encoder_embed_dim)

        # --------------------------------------------------------------------------
        # MAR decoder specifics
        self.decoder_embed = nn.Linear(encoder_embed_dim, decoder_embed_dim, bias=True)
        self.mask_token = nn.Parameter(torch.zeros(1, 1, decoder_embed_dim))
        self.decoder_pos_embed_learned = nn.Parameter(torch.zeros(1, self.seq_len + self.buffer_size, decoder_embed_dim))

        self.decoder_blocks = nn.ModuleList([
            Block(decoder_embed_dim, decoder_num_heads, mlp_ratio, qkv_bias=True, norm_layer=norm_layer,
                  proj_drop=proj_dropout, attn_drop=attn_dropout) for _ in range(decoder_depth)])

        self.decoder_norm = norm_layer(decoder_embed_dim)
        self.diffusion_pos_embed_learned = nn.Parameter(torch.zeros(1, self.seq_len, decoder_embed_dim))

        self.initialize_weights()

        # --------------------------------------------------------------------------
        # Diffusion Loss
        self.diffloss = DiffLoss(
            target_channels=self.token_embed_dim,
            z_channels=decoder_embed_dim,
            width=diffloss_w,
            depth=diffloss_d,
            num_sampling_steps=num_sampling_steps,
            grad_checkpointing=grad_checkpointing
        )
        self.diffusion_batch_mul = diffusion_batch_mul

    def initialize_weights(self):
        # parameters
        torch.nn.init.normal_(self.class_emb.weight, std=.02)
        torch.nn.init.normal_(self.fake_latent, std=.02)
        if self.use_dataset_conditioning:
            torch.nn.init.normal_(self.dataset_emb.weight, std=.02)
            torch.nn.init.normal_(self.fake_dataset_latent, std=.02)
        torch.nn.init.normal_(self.mask_token, std=.02)
        torch.nn.init.normal_(self.encoder_pos_embed_learned, std=.02)
        torch.nn.init.normal_(self.decoder_pos_embed_learned, std=.02)
        torch.nn.init.normal_(self.diffusion_pos_embed_learned, std=.02)

        # initialize nn.Linear and nn.LayerNorm
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            # we use xavier_uniform following official JAX ViT:
            torch.nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)
            if m.weight is not None:
                nn.init.constant_(m.weight, 1.0)

    def patchify(self, x):
        if len(x.shape) == 5: 
            x = x[:, :, 0, ...]
        bsz, c, h, w = x.shape
        p = self.patch_size
        h_, w_ = h // p, w // p

        x = x.reshape(bsz, c, h_, p, w_, p)
        x = torch.einsum('nchpwq->nhwcpq', x)
        x = x.reshape(bsz, h_ * w_, c * p ** 2)
        return x  # [n, l, d]

    def unpatchify(self, x):
        bsz = x.shape[0]
        p = self.patch_size
        c = self.in_channels
        h_, w_ = self.seq_h, self.seq_w

        x = x.reshape(bsz, h_, w_, c, p, p)
        x = torch.einsum('nhwcpq->nchpwq', x)
        x = x.reshape(bsz, c, h_ * p, w_ * p)
        return x  # [n, c, h, w]

    def sample_orders(self, bsz):
        # generate a batch of random generation orders
        orders = []
        for _ in range(bsz):
            order = np.array(list(range(self.seq_len)))
            np.random.shuffle(order)
            orders.append(order)
        orders = torch.Tensor(np.array(orders)).cuda().long()
        return orders

    def random_masking(self, x, orders):
        # generate token mask
        bsz, seq_len, embed_dim = x.shape
        mask_rate = self.mask_ratio_generator.rvs(1)[0]
        num_masked_tokens = int(np.ceil(seq_len * mask_rate))
        mask = torch.zeros(bsz, seq_len, device=x.device)
        mask = torch.scatter(mask, dim=-1, index=orders[:, :num_masked_tokens],
                             src=torch.ones(bsz, seq_len, device=x.device))
        return mask

    def forward_mae_encoder(self, x, mask, class_embedding, dataset_embedding=None):
        x = self.z_proj(x)
        bsz, seq_len, embed_dim = x.shape

        # concat buffer
        x = torch.cat([torch.zeros(bsz, self.buffer_size, embed_dim, device=x.device), x], dim=1)
        mask_with_buffer = torch.cat([torch.zeros(x.size(0), self.buffer_size, device=x.device), mask], dim=1)

        # random drop class embedding during training
        if self.training:
            drop_latent_mask = torch.rand(bsz) < self.label_drop_prob
            drop_latent_mask = drop_latent_mask.unsqueeze(-1).cuda().to(x.dtype)
            class_embedding = drop_latent_mask * self.fake_latent + (1 - drop_latent_mask) * class_embedding

        # combine class and dataset embeddings
        combined_embedding = class_embedding
        if dataset_embedding is not None:
            combined_embedding = combined_embedding + dataset_embedding

        x[:, :self.buffer_size] = combined_embedding.unsqueeze(1)

        # encoder position embedding
        x = x + self.encoder_pos_embed_learned
        x = self.z_proj_ln(x)

        # dropping
        x = x[(1-mask_with_buffer).nonzero(as_tuple=True)].reshape(bsz, -1, embed_dim)

        # apply Transformer blocks
        if self.grad_checkpointing and not torch.jit.is_scripting():
            for block in self.encoder_blocks:
                x = checkpoint(block, x)
        else:
            for block in self.encoder_blocks:
                x = block(x)
        x = self.encoder_norm(x)

        return x

    def forward_mae_decoder(self, x, mask):

        x = self.decoder_embed(x)
        mask_with_buffer = torch.cat([torch.zeros(x.size(0), self.buffer_size, device=x.device), mask], dim=1)

        # pad mask tokens
        mask_tokens = self.mask_token.repeat(mask_with_buffer.shape[0], mask_with_buffer.shape[1], 1).to(x.dtype)
        x_after_pad = mask_tokens.clone()
        x_after_pad[(1 - mask_with_buffer).nonzero(as_tuple=True)] = x.reshape(x.shape[0] * x.shape[1], x.shape[2])

        # decoder position embedding
        x = x_after_pad + self.decoder_pos_embed_learned

        # apply Transformer blocks
        if self.grad_checkpointing and not torch.jit.is_scripting():
            for block in self.decoder_blocks:
                x = checkpoint(block, x)
        else:
            for block in self.decoder_blocks:
                x = block(x)
        x = self.decoder_norm(x)

        x = x[:, self.buffer_size:]
        x = x + self.diffusion_pos_embed_learned
        return x

    def forward_loss(self, z, target, mask):
        loss = self.diffloss(z=z, target=target, mask=mask)
        return loss

    def forward(self, imgs, y=None, dataset_id=None):

        # class embed
        class_embedding = self.class_emb(y)
        
        # dataset embed (optional)
        dataset_embedding = None
        if self.use_dataset_conditioning and dataset_id is not None:
            dataset_embedding = self.dataset_emb(dataset_id)

        # patchify and mask (drop) tokens
        x = self.patchify(imgs)
        gt_latents = x.clone().detach()
        orders = self.sample_orders(bsz=x.size(0))
        mask = self.random_masking(x, orders)

        # mae encoder
        x = self.forward_mae_encoder(x, mask, class_embedding, dataset_embedding)

        # mae decoder
        z = self.forward_mae_decoder(x, mask)

        # diffloss
        loss = self.forward_loss(z=z, target=gt_latents, mask=mask)

        return loss

    def sample(self, bsz, device, *args, tokens=None, mask=None, orders=None, num_iter=16, cfg=1.0, cfg_schedule="linear", y=None, dataset_id=None, temperature=1.0, progress=False, **kwargs):

        mask = torch.ones(bsz, self.seq_len).to(device) if mask is None else mask
        tokens = torch.zeros(bsz, self.seq_len, self.token_embed_dim).to(device) if tokens is None else tokens
        orders = self.sample_orders(bsz).to(device) if orders is None else orders 


        indices = list(range(num_iter))
        if progress:
            indices = tqdm(indices)
        # generate latents
        for step in indices:
            cur_tokens = tokens.clone()

            # class embedding and CFG
            if y is not None:
                class_embedding = self.class_emb(y)
            else:
                class_embedding = self.fake_latent.repeat(bsz, 1)
            
            # dataset embedding (optional)
            dataset_embedding = None
            if self.use_dataset_conditioning:
                if dataset_id is not None:
                    dataset_embedding = self.dataset_emb(dataset_id)
                else:
                    dataset_embedding = self.fake_dataset_latent.repeat(bsz, 1)
            
            if not cfg == 1.0:
                tokens = torch.cat([tokens, tokens], dim=0)
                class_embedding = torch.cat([class_embedding, self.fake_latent.repeat(bsz, 1)], dim=0)
                if dataset_embedding is not None:
                    dataset_embedding = torch.cat([dataset_embedding, self.fake_dataset_latent.repeat(bsz, 1)], dim=0)
                mask = torch.cat([mask, mask], dim=0)

            # mae encoder
            x = self.forward_mae_encoder(tokens, mask, class_embedding, dataset_embedding)

            # mae decoder
            z = self.forward_mae_decoder(x, mask)

            # mask ratio for the next round, following MaskGIT and MAGE.
            mask_ratio = np.cos(math.pi / 2. * (step + 1) / num_iter)
            mask_len = torch.Tensor([np.floor(self.seq_len * mask_ratio)]).cuda()

            # masks out at least one for the next iteration
            mask_len = torch.maximum(torch.Tensor([1]).cuda(),
                                     torch.minimum(torch.sum(mask, dim=-1, keepdims=True) - 1, mask_len))

            # get masking for next iteration and locations to be predicted in this iteration
            mask_next = mask_by_order(mask_len[0], orders, bsz, self.seq_len)
            if step >= num_iter - 1:
                mask_to_pred = mask[:bsz].bool()
            else:
                mask_to_pred = torch.logical_xor(mask[:bsz].bool(), mask_next.bool())
            mask = mask_next
            if not cfg == 1.0:
                mask_to_pred = torch.cat([mask_to_pred, mask_to_pred], dim=0)

            # sample token latents for this step
            z = z[mask_to_pred.nonzero(as_tuple=True)]
            # cfg schedule follow Muse
            if cfg_schedule == "linear":
                cfg_iter = 1 + (cfg - 1) * (self.seq_len - mask_len[0]) / self.seq_len
            elif cfg_schedule == "constant":
                cfg_iter = cfg
            else:
                raise NotImplementedError
            sampled_token_latent = self.diffloss.sample(z, temperature, cfg_iter)
            if not cfg == 1.0:
                sampled_token_latent, _ = sampled_token_latent.chunk(2, dim=0)  # Remove null class samples
                mask_to_pred, _ = mask_to_pred.chunk(2, dim=0)

            cur_tokens[mask_to_pred.nonzero(as_tuple=True)] = sampled_token_latent
            tokens = cur_tokens.clone()

        # unpatchify
        imgs = self.unpatchify(tokens)
        return imgs




from .first_stage import *  # noqa: F401,F403
from .generators import *  # noqa: F401,F403
from .modules.wrapper import GenWrapper
from .registry import (
    MODEL_REGISTRY,
    ModelInfo,
    available_models,
    get_model,
    get_model_info,
    register_model,
)

__all__ = [
    "MODEL_REGISTRY",
    "ModelInfo",
    "available_models",
    "get_model",
    "get_model_info",
    "register_model",
    "GenWrapper",
]


from functools import partial
from typing import Optional

import torch.nn as nn

from ..utils.network import build_kwargs_from_config

__all__ = ["build_act"]


# register activation function here
REGISTERED_ACT_DICT: dict[str, type] = {
    "relu": nn.ReLU,
    "relu6": nn.ReLU6,
    "hswish": nn.Hardswish,
    "silu": nn.SiLU,
    "gelu": partial(nn.GELU, approximate="tanh"),
}


def build_act(name: str, **kwargs) -> Optional[nn.Module]:
    if name in REGISTERED_ACT_DICT:
        act_cls = REGISTERED_ACT_DICT[name]
        args = build_kwargs_from_config(kwargs, act_cls)
        return act_cls(**args)
    else:
        return None



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Iterable, Optional


@dataclass(slots=True)
class ModelInfo:
    """Display-only metadata for a registered model (no builder)."""

    name: str
    description: Optional[str] = None
    code_url: Optional[str] = None
    paper_url: Optional[str] = None
    ckpt_path: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)

    def __str__(self) -> str:
        lines = [f"Model: {self.name}"]
        if self.description:
            lines.append(f"  description: {self.description}")
        if self.code_url:
            lines.append(f"  code_url: {self.code_url}")
        if self.paper_url:
            lines.append(f"  paper_url: {self.paper_url}")
        if self.ckpt_path:
            lines.append(f"  ckpt_path: {self.ckpt_path}")
        if self.metadata:
            lines.append(f"  metadata: {self.metadata}")
        return "\n".join(lines)


@dataclass(slots=True)
class ModelEntry:
    """Metadata container for a registered model."""

    name: str
    builder: Callable[..., Any]
    code_url: Optional[str] = None
    description: Optional[str] = None
    paper_url: Optional[str] = None
    ckpt_path: Optional[str] = None  # Path to original model weights
    metadata: Dict[str, Any] = field(default_factory=dict)

    def instantiate(self, *args: Any, **kwargs: Any) -> Any:
        return self.builder(*args, **kwargs)

    def to_info(self) -> ModelInfo:
        """Return display-only info (no builder)."""
        return ModelInfo(
            name=self.name,
            description=self.description,
            code_url=self.code_url,
            paper_url=self.paper_url,
            ckpt_path=self.ckpt_path,
            metadata=dict(self.metadata),
        )


class ModelRegistry:
    """Central registry that keeps track of model builder callables."""

    def __init__(self) -> None:
        self._registry: Dict[str, ModelEntry] = {}

    @staticmethod
    def _normalize(name: str) -> str:
        return name.strip().lower()

    def register(
        self,
        name: str,
        builder: Callable[..., Any],
        *,
        metadata: Optional[Dict[str, Any]] = None,
        code_url: Optional[str] = None,
        description: Optional[str] = None,
        paper_url: Optional[str] = None,
        ckpt_path: Optional[str] = None,
        override: bool = False,
    ) -> ModelEntry:
        key = self._normalize(name)
        if not override and key in self._registry:
            raise ValueError(f"Model '{name}' already registered.")
        entry = ModelEntry(
            name=key,
            builder=builder,
            code_url=code_url,
            description=description,
            paper_url=paper_url,
            ckpt_path=ckpt_path,
            metadata=metadata or {},
        )
        self._registry[key] = entry
        return entry

    def get(self, name: str) -> ModelEntry:
        key = self._normalize(name)
        try:
            return self._registry[key]
        except KeyError as exc:
            raise KeyError(
                f"Unknown model '{name}'. Available models: {sorted(self._registry)}"
            ) from exc

    def get_info(self, name: str) -> ModelInfo:
        """Return display-only metadata for a model (no builder)."""
        return self.get(name).to_info()

    def create(self, name: str, *args: Any, **kwargs: Any) -> Any:
        return self.get(name).instantiate(*args, **kwargs)

    def available(self, prefix: Optional[str] = None) -> Iterable[str]:
        if prefix is None:
            return tuple(sorted(self._registry))
        normalized = self._normalize(prefix)
        return tuple(sorted(name for name in self._registry if name.startswith(normalized)))


MODEL_REGISTRY = ModelRegistry()


def register_model(
    name: str,
    builder: Optional[Callable[..., Any]] = None,
    *,
    metadata: Optional[Dict[str, Any]] = None,
    code_url: Optional[str] = None,
    description: Optional[str] = None,
    paper_url: Optional[str] = None,
    ckpt_path: Optional[str] = None,
    override: bool = False,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    Helper that registers a model builder either directly or as a decorator.
    """

    def decorator(fn: Callable[..., Any]) -> Callable[..., Any]:
        MODEL_REGISTRY.register(
            name,
            fn,
            metadata=metadata,
            code_url=code_url,
            description=description,
            paper_url=paper_url,
            ckpt_path=ckpt_path,
            override=override,
        )
        return fn

    if builder is not None:
        return decorator(builder)
    return decorator


def get_model(name: str, *args: Any, **kwargs: Any) -> Any:
    """Instantiate a registered model."""
    return MODEL_REGISTRY.create(name, *args, **kwargs)


def get_model_info(name: str) -> ModelInfo:
    """Return display-only metadata for a registered model (no builder)."""
    return MODEL_REGISTRY.get_info(name)


def available_models(prefix: Optional[str] = None) -> Iterable[str]:
    """List the registered model identifiers."""
    return MODEL_REGISTRY.available(prefix=prefix)




# Adopted from LDM's KL-VAE: https://github.com/CompVis/latent-diffusion
import torch
import torch.nn as nn
import numpy as np


def get_conv_layer(dims):
    if dims == 2:
        return nn.Conv2d
    elif dims == 3:
        return nn.Conv3d
    else:
        raise ValueError(f"Unsupported number of dimensions: {dims}")


def get_interpolate_mode(dims):
    if dims == 2:
        return "nearest"
    elif dims == 3:
        return "nearest"
    else:
        raise ValueError(f"Unsupported number of dimensions: {dims}")


def nonlinearity(x):
    # swish
    return x * torch.sigmoid(x)


def Normalize(in_channels, num_groups=32):
    return torch.nn.GroupNorm(
        num_groups=num_groups, num_channels=in_channels, eps=1e-6, affine=True
    )


class Upsample(nn.Module):
    def __init__(self, in_channels, with_conv, dims=2):
        super().__init__()
        self.with_conv = with_conv
        self.dims = dims
        if self.with_conv:
            conv_layer = get_conv_layer(dims)
            kernel_size = 3
            self.conv = conv_layer(
                in_channels, in_channels, kernel_size=kernel_size, stride=1, padding=1
            )

    def forward(self, x):
        x = torch.nn.functional.interpolate(x, scale_factor=2.0, mode=get_interpolate_mode(self.dims))
        if self.with_conv:
            x = self.conv(x)
        return x


class Downsample(nn.Module):
    def __init__(self, in_channels, with_conv, dims=2):
        super().__init__()
        self.with_conv = with_conv
        self.dims = dims
        if self.with_conv:
            conv_layer = get_conv_layer(dims)
            kernel_size = 3
            self.conv = conv_layer(
                in_channels, in_channels, kernel_size=kernel_size, stride=2, padding=0
            )

    def forward(self, x):
        if self.with_conv:
            pad_size = (0, 1) * self.dims  # Creates appropriate padding tuple for 2D/3D
            x = torch.nn.functional.pad(x, pad_size, mode="constant", value=0)
            x = self.conv(x)
        else:
            if self.dims == 2:
                x = torch.nn.functional.avg_pool2d(x, kernel_size=2, stride=2)
            else:
                x = torch.nn.functional.avg_pool3d(x, kernel_size=2, stride=2)
        return x


class ResnetBlock(nn.Module):
    def __init__(
        self,
        *,
        in_channels,
        out_channels=None,
        conv_shortcut=False,
        dropout,
        temb_channels=512,
        dims=2
    ):
        super().__init__()
        self.in_channels = in_channels
        out_channels = in_channels if out_channels is None else out_channels
        self.out_channels = out_channels
        self.use_conv_shortcut = conv_shortcut
        conv_layer = get_conv_layer(dims)

        self.norm1 = Normalize(in_channels)
        self.conv1 = conv_layer(
            in_channels, out_channels, kernel_size=3, stride=1, padding=1
        )
        if temb_channels > 0:
            self.temb_proj = torch.nn.Linear(temb_channels, out_channels)
        self.norm2 = Normalize(out_channels)
        self.dropout = torch.nn.Dropout(dropout)
        self.conv2 = conv_layer(
            out_channels, out_channels, kernel_size=3, stride=1, padding=1
        )
        if self.in_channels != self.out_channels:
            if self.use_conv_shortcut:
                self.conv_shortcut = conv_layer(
                    in_channels, out_channels, kernel_size=3, stride=1, padding=1
                )
            else:
                self.nin_shortcut = conv_layer(
                    in_channels, out_channels, kernel_size=1, stride=1, padding=0
                )

    def forward(self, x, temb):
        h = x
        h = self.norm1(h)
        h = nonlinearity(h)
        h = self.conv1(h)

        if temb is not None:
            h = h + self.temb_proj(nonlinearity(temb))[:, :, None, None]

        h = self.norm2(h)
        h = nonlinearity(h)
        h = self.dropout(h)
        h = self.conv2(h)

        if self.in_channels != self.out_channels:
            if self.use_conv_shortcut:
                x = self.conv_shortcut(x)
            else:
                x = self.nin_shortcut(x)

        return x + h


class AttnBlock(nn.Module):
    def __init__(self, in_channels, dims=2):
        super().__init__()
        self.in_channels = in_channels
        self.dims = dims
        conv_layer = get_conv_layer(dims)

        self.norm = Normalize(in_channels)
        self.q = conv_layer(
            in_channels, in_channels, kernel_size=1, stride=1, padding=0
        )
        self.k = conv_layer(
            in_channels, in_channels, kernel_size=1, stride=1, padding=0
        )
        self.v = conv_layer(
            in_channels, in_channels, kernel_size=1, stride=1, padding=0
        )
        self.proj_out = conv_layer(
            in_channels, in_channels, kernel_size=1, stride=1, padding=0
        )

    def forward(self, x):
        h_ = x
        h_ = self.norm(h_)
        q = self.q(h_)
        k = self.k(h_)
        v = self.v(h_)

        # compute attention
        if self.dims == 2:
            b, c, h, w = q.shape
            q = q.reshape(b, c, h * w)
            k = k.reshape(b, c, h * w)
            v = v.reshape(b, c, h * w)
        else:  # dims == 3
            b, c, d, h, w = q.shape
            q = q.reshape(b, c, d * h * w)
            k = k.reshape(b, c, d * h * w)
            v = v.reshape(b, c, d * h * w)

        q = q.permute(0, 2, 1)  # b,hw,c
        w_ = torch.bmm(q, k)  # b,hw,hw
        w_ = w_ * (int(c) ** (-0.5))
        w_ = torch.nn.functional.softmax(w_, dim=2)

        # attend to values
        w_ = w_.permute(0, 2, 1)
        h_ = torch.bmm(v, w_)

        if self.dims == 2:
            h_ = h_.reshape(b, c, h, w)
        else:  # dims == 3
            h_ = h_.reshape(b, c, d, h, w)

        h_ = self.proj_out(h_)

        return x + h_


class Encoder(nn.Module):
    def __init__(
        self,
        *,
        ch=128,
        out_ch=3,
        ch_mult=(1, 1, 2, 2, 4),
        num_res_blocks=2,
        attn_resolutions=(16,),
        dropout=0.0,
        resamp_with_conv=True,
        in_channels=3,
        img_size=256,
        z_channels=16,
        double_z=True,
        dims=2,
        ignore_mid_attn=False,
        **ignore_kwargs,
    ):
        super().__init__()
        self.ch = ch
        self.temb_ch = 0
        self.num_resolutions = len(ch_mult)
        self.vae_stride = 2 ** (len(ch_mult) - 1)
        self.num_res_blocks = num_res_blocks
        
        # Handle img_size: support both int (backward compat) and tuple/list (non-cubic)
        if isinstance(img_size, (int, float)):
            # Backward compatibility: convert int to tuple
            if dims == 2:
                self.resolution = (int(img_size),) * 2
            else:  # dims == 3
                self.resolution = (int(img_size),) * 3
        elif isinstance(img_size, (tuple, list)):
            self.resolution = tuple(int(s) for s in img_size)
            if len(self.resolution) != dims:
                raise ValueError(f"img_size tuple length {len(self.resolution)} doesn't match dims {dims}")
        else:
            raise ValueError(f"img_size must be int or tuple/list, got {type(img_size)}")
        
        self.in_channels = in_channels
        self.dims = dims
        self.ignore_mid_attn = ignore_mid_attn
        self.z_channels = z_channels

        conv_layer = get_conv_layer(dims)

        # downsampling
        self.conv_in = conv_layer(
            in_channels, self.ch, kernel_size=3, stride=1, padding=1
        )

        curr_res = list(self.resolution)  # Track resolution as list for easy modification
        in_ch_mult = (1,) + tuple(ch_mult)
        self.down = nn.ModuleList()
        for i_level in range(self.num_resolutions):
            block = nn.ModuleList()
            attn = nn.ModuleList()
            block_in = ch * in_ch_mult[i_level]
            block_out = ch * ch_mult[i_level]
            for i_block in range(self.num_res_blocks):
                block.append(
                    ResnetBlock(
                        in_channels=block_in,
                        out_channels=block_out,
                        temb_channels=self.temb_ch,
                        dropout=dropout,
                        dims=dims
                    )
                )
                block_in = block_out
                # Check if any dimension matches attn_resolutions
                if any(res in attn_resolutions for res in curr_res):
                    attn.append(AttnBlock(block_in, dims=dims))
            down = nn.Module()
            down.block = block
            down.attn = attn
            if i_level != self.num_resolutions - 1:
                down.downsample = Downsample(block_in, resamp_with_conv, dims=dims)
                # Downsample each dimension
                curr_res = [r // 2 for r in curr_res]
            self.down.append(down)

        # middle
        self.mid = nn.Module()
        self.mid.block_1 = ResnetBlock(
            in_channels=block_in,
            out_channels=block_in,
            temb_channels=self.temb_ch,
            dropout=dropout,
            dims=dims
        )
        if not ignore_mid_attn:
            self.mid.attn_1 = AttnBlock(block_in, dims=dims)
        self.mid.block_2 = ResnetBlock(
            in_channels=block_in,
            out_channels=block_in,
            temb_channels=self.temb_ch,
            dropout=dropout,
            dims=dims
        )

        # end
        self.norm_out = Normalize(block_in)
        self.conv_out = conv_layer(
            block_in,
            2 * z_channels if double_z else z_channels,
            kernel_size=3,
            stride=1,
            padding=1,
        )

    def forward(self, x):
        # assert x.shape[2] == x.shape[3] == self.resolution, "{}, {}, {}".format(x.shape[2], x.shape[3], self.resolution)

        # timestep embedding
        temb = None

        # downsampling
        hs = [self.conv_in(x)]
        for i_level in range(self.num_resolutions):
            for i_block in range(self.num_res_blocks):
                h = self.down[i_level].block[i_block](hs[-1], temb)
                if len(self.down[i_level].attn) > 0:
                    h = self.down[i_level].attn[i_block](h)
                hs.append(h)
            if i_level != self.num_resolutions - 1:
                hs.append(self.down[i_level].downsample(hs[-1]))

        # middle
        h = hs[-1]
        h = self.mid.block_1(h, temb)
        if not self.ignore_mid_attn:
            h = self.mid.attn_1(h)
        h = self.mid.block_2(h, temb)

        # end
        h = self.norm_out(h)
        h = nonlinearity(h)
        h = self.conv_out(h)
        return h


class Decoder(nn.Module):
    def __init__(
        self,
        *,
        ch=128,
        out_ch=3,
        ch_mult=(1, 1, 2, 2, 4),
        num_res_blocks=2,
        attn_resolutions=(),
        dropout=0.0,
        resamp_with_conv=True,
        in_channels=3,
        img_size=256,
        z_channels=16,
        give_pre_end=False,
        dims=2,
        ignore_mid_attn=False,
        **ignore_kwargs,
    ):
        super().__init__()
        self.ch = ch
        self.temb_ch = 0
        self.num_resolutions = len(ch_mult)
        self.num_res_blocks = num_res_blocks
        
        # Handle img_size: support both int (backward compat) and tuple/list (non-cubic)
        if isinstance(img_size, (int, float)):
            # Backward compatibility: convert int to tuple
            if dims == 2:
                self.resolution = (int(img_size),) * 2
            else:  # dims == 3
                self.resolution = (int(img_size),) * 3
        elif isinstance(img_size, (tuple, list)):
            self.resolution = tuple(int(s) for s in img_size)
            if len(self.resolution) != dims:
                raise ValueError(f"img_size tuple length {len(self.resolution)} doesn't match dims {dims}")
        else:
            raise ValueError(f"img_size must be int or tuple/list, got {type(img_size)}")
        
        self.in_channels = in_channels
        self.give_pre_end = give_pre_end
        self.dims = dims
        self.ignore_mid_attn = ignore_mid_attn
        conv_layer = get_conv_layer(dims)

        # compute in_ch_mult, block_in and curr_res at lowest res
        in_ch_mult = (1,) + tuple(ch_mult)
        block_in = ch * ch_mult[self.num_resolutions - 1]
        # Calculate curr_res for each dimension
        curr_res = [r // (2 ** (self.num_resolutions - 1)) for r in self.resolution]
        
        # Adjust z_shape based on dimensions (non-cubic support)
        if dims == 2:
            self.z_shape = (1, z_channels, curr_res[0], curr_res[1])
        else:  # dims == 3
            self.z_shape = (1, z_channels, curr_res[0], curr_res[1], curr_res[2])
            
        print(
            "Working with z of shape {} = {} dimensions.".format(
                self.z_shape, np.prod(self.z_shape)
            )
        )

        # z to block_in
        self.conv_in = conv_layer(
            z_channels, block_in, kernel_size=3, stride=1, padding=1
        )

        # middle
        self.mid = nn.Module()
        self.mid.block_1 = ResnetBlock(
            in_channels=block_in,
            out_channels=block_in,
            temb_channels=self.temb_ch,
            dropout=dropout,
            dims=dims
        )
        if not ignore_mid_attn:
            self.mid.attn_1 = AttnBlock(block_in, dims=dims)
        self.mid.block_2 = ResnetBlock(
            in_channels=block_in,
            out_channels=block_in,
            temb_channels=self.temb_ch,
            dropout=dropout,
            dims=dims
        )

        # upsampling
        self.up = nn.ModuleList()
        for i_level in reversed(range(self.num_resolutions)):
            block = nn.ModuleList()
            attn = nn.ModuleList()
            block_out = ch * ch_mult[i_level]
            for i_block in range(self.num_res_blocks + 1):
                block.append(
                    ResnetBlock(
                        in_channels=block_in,
                        out_channels=block_out,
                        temb_channels=self.temb_ch,
                        dropout=dropout,
                        dims=dims
                    )
                )
                block_in = block_out
                # Check if any dimension matches attn_resolutions
                if any(res in attn_resolutions for res in curr_res):
                    attn.append(AttnBlock(block_in, dims=dims))
            up = nn.Module()
            up.block = block
            up.attn = attn
            if i_level != 0:
                up.upsample = Upsample(block_in, resamp_with_conv, dims=dims)
                # Upsample each dimension
                curr_res = [r * 2 for r in curr_res]
            self.up.insert(0, up)  # prepend to get consistent order

        # end
        self.norm_out = Normalize(block_in)
        self.conv_out = conv_layer(
            block_in, out_ch, kernel_size=3, stride=1, padding=1
        )

    def forward(self, z):
        # assert z.shape[1:] == self.z_shape[1:]
        self.last_z_shape = z.shape

        # timestep embedding
        temb = None

        # z to block_in
        h = self.conv_in(z)

        # middle
        h = self.mid.block_1(h, temb)
        if not self.ignore_mid_attn:
            h = self.mid.attn_1(h)
        h = self.mid.block_2(h, temb)

        # upsampling
        for i_level in reversed(range(self.num_resolutions)):
            for i_block in range(self.num_res_blocks + 1):
                h = self.up[i_level].block[i_block](h, temb)
                if len(self.up[i_level].attn) > 0:
                    h = self.up[i_level].attn[i_block](h)
            if i_level != 0:
                h = self.up[i_level].upsample(h)

        # end
        if self.give_pre_end:
            return h

        h = self.norm_out(h)
        h = nonlinearity(h)
        h = self.conv_out(h)
        return h

    def get_last_layer(self) -> nn.Parameter:
        return self.conv_out.weight

### ViTVQ ### 

def init_weights(m):
    if isinstance(m, nn.Linear):
        torch.nn.init.xavier_uniform_(m.weight)
        if m.bias is not None:
            nn.init.constant_(m.bias, 0)
    elif isinstance(m, nn.LayerNorm):
        nn.init.constant_(m.bias, 0)
        nn.init.constant_(m.weight, 1.0)
    elif isinstance(m, (nn.Conv2d, nn.ConvTranspose2d, nn.Conv3d, nn.ConvTranspose3d)):
        w = m.weight.data
        torch.nn.init.xavier_uniform_(w.view([w.shape[0], -1]))



# modelling/alignments.py
from abc import ABC, abstractmethod
from typing import Optional, Tuple
import torch
import torch.nn as nn
import torch.nn.functional as F
import math
import cv2
import numpy as np
import open_clip
import torchvision.transforms as T

# For external models (DINO, CLIP) - try importing timm but make it optional
try:
    from timm import create_model
    TIMM_AVAILABLE = True
except ImportError:
    TIMM_AVAILABLE = False
    create_model = None




class HOGGenerator(nn.Module):
    """Generate HOG feature for images.

    This module is used in MaskFeat to generate HOG feature. The code is
    modified from file `slowfast/models/operators.py
    <https://github.com/facebookresearch/SlowFast/blob/main/slowfast/models/operators.py>`_.
    Here is the link of `HOG wikipedia
    <https://en.wikipedia.org/wiki/Histogram_of_oriented_gradients>`_.

    Args:
        nbins (int): Number of bin. Defaults to 9.
        pool (float): Number of cell. Defaults to 8.
        gaussian_window (int): Size of gaussian kernel. Defaults to 16.
    """

    def __init__(self,
                 nbins: int = 9,
                 pool: int = 8,
                 gaussian_window: int = 16) -> None:
        super().__init__()
        self.nbins = nbins
        self.pool = pool
        self.pi = math.pi
        weight_x = torch.FloatTensor([[1, 0, -1], [2, 0, -2], [1, 0, -1]])
        weight_x = weight_x.view(1, 1, 3, 3).repeat(3, 1, 1, 1).contiguous()
        weight_y = weight_x.transpose(2, 3).contiguous()
        self.register_buffer('weight_x', weight_x)
        self.register_buffer('weight_y', weight_y)

        self.gaussian_window = gaussian_window
        if gaussian_window:
            gaussian_kernel = self.get_gaussian_kernel(gaussian_window,
                                                       gaussian_window // 2)
            self.register_buffer('gaussian_kernel', gaussian_kernel)

    def get_gaussian_kernel(self, kernlen: int, std: int) -> torch.Tensor:
        """Returns a 2D Gaussian kernel array."""

        def _gaussian_fn(kernlen: int, std: int) -> torch.Tensor:
            n = torch.arange(0, kernlen).float()
            n -= n.mean()
            n /= std
            w = torch.exp(-0.5 * n**2)
            return w

        kernel_1d = _gaussian_fn(kernlen, std)
        kernel_2d = kernel_1d[:, None] * kernel_1d[None, :]
        return kernel_2d / kernel_2d.sum()

    def _reshape(self, hog_feat: torch.Tensor) -> torch.Tensor:
        """Reshape HOG Features for output."""
        hog_feat = hog_feat.flatten(1, 2)
        self.unfold_size = hog_feat.shape[-1] // 16
        hog_feat = hog_feat.permute(0, 2, 3, 1)
        # print(hog_feat.shape)
        hog_feat = hog_feat.unfold(1, self.unfold_size,
                                   self.unfold_size).unfold(
                                       2, self.unfold_size, self.unfold_size)
        hog_feat = hog_feat.flatten(1, 2).flatten(2)
        return hog_feat

    @torch.no_grad()
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Generate hog feature for each batch images.

        Args:
            x (torch.Tensor): Input images of shape (N, 3, H, W).

        Returns:
            torch.Tensor: Hog features.
        """
        # input is RGB image with shape [B 3 H W]
        self.h, self.w = x.size(-2), x.size(-1)
        x = F.pad(x, pad=(1, 1, 1, 1), mode='reflect')
        gx_rgb = F.conv2d(
            x, self.weight_x, bias=None, stride=1, padding=0, groups=3)
        gy_rgb = F.conv2d(
            x, self.weight_y, bias=None, stride=1, padding=0, groups=3)
        norm_rgb = torch.stack([gx_rgb, gy_rgb], dim=-1).norm(dim=-1)
        phase = torch.atan2(gx_rgb, gy_rgb)
        phase = phase / self.pi * self.nbins  # [-9, 9]

        b, c, h, w = norm_rgb.shape
        out = torch.zeros((b, c, self.nbins, h, w),
                          dtype=torch.float,
                          device=x.device)
        phase = phase.view(b, c, 1, h, w)
        norm_rgb = norm_rgb.view(b, c, 1, h, w)
        if self.gaussian_window:
            if h != self.gaussian_window:
                assert h % self.gaussian_window == 0, 'h {} gw {}'.format(
                    h, self.gaussian_window)
                repeat_rate = h // self.gaussian_window
                temp_gaussian_kernel = self.gaussian_kernel.repeat(
                    [repeat_rate, repeat_rate])
            else:
                temp_gaussian_kernel = self.gaussian_kernel
            norm_rgb *= temp_gaussian_kernel

        out.scatter_add_(2, phase.floor().long() % self.nbins, norm_rgb)

        out = out.unfold(3, self.pool, self.pool)
        out = out.unfold(4, self.pool, self.pool)
        out = out.sum(dim=[-1, -2])

        self.out = F.normalize(out, p=2, dim=2)

        return self._reshape(self.out)

    def generate_hog_image(self, hog_out: torch.Tensor) -> np.ndarray:
        """Generate HOG image according to HOG features."""
        assert hog_out.size(0) == 1 and hog_out.size(1) == 3, \
            'Check the input batch size and the channcel number, only support'\
            '"batch_size = 1".'
        hog_image = np.zeros([self.h, self.w])
        cell_gradient = np.array(hog_out.mean(dim=1).squeeze().detach().cpu())
        cell_width = self.pool / 2
        max_mag = np.array(cell_gradient).max()
        angle_gap = 360 / self.nbins

        for x in range(cell_gradient.shape[1]):
            for y in range(cell_gradient.shape[2]):
                cell_grad = cell_gradient[:, x, y]
                cell_grad /= max_mag
                angle = 0
                for magnitude in cell_grad:
                    angle_radian = math.radians(angle)
                    x1 = int(x * self.pool +
                             magnitude * cell_width * math.cos(angle_radian))
                    y1 = int(y * self.pool +
                             magnitude * cell_width * math.sin(angle_radian))
                    x2 = int(x * self.pool -
                             magnitude * cell_width * math.cos(angle_radian))
                    y2 = int(y * self.pool -
                             magnitude * cell_width * math.sin(angle_radian))
                    magnitude = 0 if magnitude < 0 else magnitude
                    cv2.line(hog_image, (y1, x1), (y2, x2),
                             int(255 * math.sqrt(magnitude)))
                    angle += angle_gap
        return hog_image


def mean_flat(x):
    return torch.mean(x, dim=list(range(1, len(x.size()))))

class AlignmentModule(ABC, nn.Module):
    """
    Base class for auxiliary alignment modules.
    Each module:
      - contains a decoder (MAETokViTDecoder)
      - contains projection heads (post_quant_conv and to_pixel)
      - contains a target model (frozen or external callable)
    Subclasses must implement `compute_target` to obtain the target representation for an input image.
    """

    def __init__(self, name: str):
        super().__init__()
        self.name = name

    @abstractmethod
    def compute_target(self, x: torch.Tensor) -> torch.Tensor:
        """
        Compute the target features from the input image x.
        Should not require gradients (freeze target model).
        Returns tensor shaped (B, L, D_target)
        """
        raise NotImplementedError

    @abstractmethod
    def decode_projection(self, quant: torch.Tensor) -> torch.Tensor:
        """
        Decode quantized tokens to predicted features in target space.
        Returns tensor shaped (B, L, D_target)
        """
        raise NotImplementedError

    def forward(self, quant: torch.Tensor, input_image: Optional[torch.Tensor] = None, mask: Optional[torch.Tensor] = None) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        """
        Compute alignment loss between decoder(quant) and target features from input_image.
        Returns: (loss, predicted_features) — predicted_features optional (for logging)
        If mask is provided (same mask used in MaskAEModel), apply mask-aware reduction as in original code.
        """
        if input_image is None:
            raise ValueError("AlignmentModule requires input_image to compute target features")

        # target features (usually frozen model)
        with torch.no_grad():
            target = self.compute_target(input_image)  # (B, L, D_target)

        # Allow subclasses to adapt projection heads if target dim changes with image size
        self.ensure_projection_dim(target.size(-1))

        # predicted features from decoder/projection
        pred = self.decode_projection(quant)  # (B, L_pred, D)

        # Align token grids if lengths differ (e.g., different patch sizes)
        if pred.shape[1] != target.shape[1]:
            pred = self._interpolate_tokens_to_match(pred, target)
            if mask is not None:
                mask = self._interpolate_mask_to_match(mask, target)
        elif mask is not None and mask.shape[1] != pred.shape[1]:
            # If pred and target already match but mask length differs, align mask to pred
            mask = self._interpolate_mask_to_match(mask, pred)

        # normalize both (original code normalized for dino/clip)
        pred_n = F.normalize(pred, dim=-1)
        target_n = F.normalize(target, dim=-1)

        # compute per-token negative cosine (like original)
        per_token = -(pred_n * target_n).sum(dim=-1, keepdim=True)  # (B, L, 1)

        if mask is not None:
            # mask shape expected (B, L, 1) or (B, L)
            if mask.dim() == 2:
                mask = mask.unsqueeze(-1)
            masked_sum = (per_token * mask).sum()
            denom = mask.sum().clamp(min=1.0)
            loss = masked_sum / denom
        else:
            loss = mean_flat(per_token.squeeze(-1))
            loss = loss.mean()

        return loss, pred

    # ------------------------------------------------------------------ #
    # Helpers for dynamic projection and token/mask alignment
    # ------------------------------------------------------------------ #
    def ensure_projection_dim(self, target_dim: int):
        """Subclasses can override if their projection depends on target dim."""
        return

    def _infer_grid_hw(self, seq_len: int) -> Tuple[int, int]:
        """Infer a square grid (h, w) from sequence length."""
        side = int(math.sqrt(seq_len))
        if side * side != seq_len:
            raise ValueError(f"Cannot infer square grid from seq_len={seq_len}")
        return side, side

    def _interpolate_tokens_to_match(self, pred: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """
        Reshape pred tokens to grid, interpolate to target grid, then flatten back.
        Assumes 2D tokens.
        """
        b, lp, c = pred.shape
        lt = target.shape[1]
        ph, pw = self._infer_grid_hw(lp)
        th, tw = self._infer_grid_hw(lt)
        pred_map = pred.view(b, ph, pw, c).permute(0, 3, 1, 2)  # (B, C, H, W)
        pred_map = F.interpolate(pred_map, size=(th, tw), mode='bilinear', align_corners=False)
        pred = pred_map.permute(0, 2, 3, 1).reshape(b, lt, c)
        return pred

    def _interpolate_mask_to_match(self, mask: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """
        Reshape mask tokens to grid, interpolate to target grid, then flatten back.
        Keeps mask float-valued.
        """
        if mask.dim() == 2:
            mask = mask.unsqueeze(-1)
        b, lp, c = mask.shape
        lt = target.shape[1]
        ph, pw = self._infer_grid_hw(lp)
        th, tw = self._infer_grid_hw(lt)
        mask_map = mask.view(b, ph, pw, c).permute(0, 3, 1, 2)  # (B, C, H, W)
        mask_map = F.interpolate(mask_map, size=(th, tw), mode='nearest')
        mask = mask_map.permute(0, 2, 3, 1).reshape(b, lt, c)
        return mask

########################################################################
# HOG alignment module
########################################################################
class HOGAlignment(AlignmentModule):
    def __init__(
        self,
        decoder: nn.Module,
        codebook_embed_dim: int,
        use_movq: bool = False,
    ):
        super().__init__('hog')

        if HOGGenerator is None:
            raise RuntimeError("HOGGenerator not available; ensure modules.hog exists or pass an alternative.")

        # Use provided decoder
        self.decoder = decoder
        self.post_quant_conv = nn.Linear(codebook_embed_dim, self.decoder.embed_dim)
        # final pixel projection in original produced 108-d HOG channels
        self.to_pixel = nn.Linear(self.decoder.embed_dim, 108)
        self.hog_generator = HOGGenerator()

        self.hog_use_movq = use_movq

    def ensure_projection_dim(self, target_dim: int):
        # Rebuild projection if target channels change with image size
        if self.to_pixel.out_features != target_dim:
            self.to_pixel = nn.Linear(self.decoder.embed_dim, target_dim).to(self.to_pixel.weight.device)

    def compute_target(self, x: torch.Tensor) -> torch.Tensor:
        # HOG generator returns (B, L, 108) presumably
        z_hog = self.hog_generator(x)
        return z_hog

    def decode_projection(self, quant: torch.Tensor) -> torch.Tensor:
        tmp = quant
        x = self.post_quant_conv(quant)
        # decoder signature: decoder(z, interpolate_zq, H, W, D) 
        if self.hog_use_movq:
            dec = self.decoder(x, interpolate_zq=tmp, H=None, W=None, D=None)
        else:
            dec = self.decoder(x, interpolate_zq=None, H=None, W=None, D=None)
        # dec should be (B, L, embed_dim) when to_pixel='identity'
        # Apply to_pixel Linear layer to project to HOG feature dimension
        dec = self.to_pixel(dec)
        return dec

########################################################################
# Dino alignment module
########################################################################
class DinoAlignment(AlignmentModule):
    def __init__(
        self,
        decoder: nn.Module,
        codebook_embed_dim: int,
        img_size: int,
        repa_model_name: str = 'vit_large_patch14_dinov2.lvd142m',
        repa_patch_size: int = 14,
        use_movq: bool = False,
    ):
        super().__init__('dino')
        
        if not TIMM_AVAILABLE:
            raise RuntimeError("timm is required for DinoAlignment. Please install timm.")
        
        # Instantiate the repa/dinov2 model
        self.repa_model = create_model(repa_model_name, pretrained=True, img_size=img_size, patch_size=repa_patch_size)
        for p in self.repa_model.parameters():
            p.requires_grad = False
        self.repa_model.eval()
        
        # Normalization for DINO (ImageNet normalization)
        self.normalize = self._create_normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        self.denormalize = self._create_denormalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        
        # Use provided decoder
        self.decoder = decoder
        self.post_quant_conv = nn.Linear(codebook_embed_dim, self.decoder.embed_dim)
        # final projection to repa_model.embed_dim
        self.to_pixel = nn.Linear(self.decoder.embed_dim, self.repa_model.embed_dim)
        self.dino_use_movq = use_movq
    
    def _create_normalize(self, mean, std):
        class Normalize(nn.Module):
            def __init__(self, mean, std):
                super().__init__()
                self.register_buffer('mean', torch.tensor(mean).view(1, -1, 1, 1))
                self.register_buffer('std', torch.tensor(std).view(1, -1, 1, 1))
            def forward(self, x):
                return (x - self.mean) / self.std
        return Normalize(mean, std)
    
    def _create_denormalize(self, mean, std):
        class Denormalize(nn.Module):
            def __init__(self, mean, std):
                super().__init__()
                self.register_buffer('mean', torch.tensor(mean).view(1, -1, 1, 1))
                self.register_buffer('std', torch.tensor(std).view(1, -1, 1, 1))
            def forward(self, x):
                return x * self.std + self.mean
        return Denormalize(mean, std)

    def compute_target(self, x: torch.Tensor) -> torch.Tensor:
        # Preprocess for repa_model: normalize using ImageNet stats
        # First denormalize from [0, 1] to ImageNet range, then normalize
        x_normalized = self.normalize(self.denormalize(x))
        z = self.repa_model.forward_features(x_normalized)[:, self.repa_model.num_prefix_tokens:]
        return z

    def decode_projection(self, quant: torch.Tensor) -> torch.Tensor:
        tmp = quant
        x = self.post_quant_conv(quant)
        if self.dino_use_movq:
            dec = self.decoder(x, interpolate_zq=tmp, H=None, W=None, D=None)
        else:
            dec = self.decoder(x, interpolate_zq=None, H=None, W=None, D=None)
        dec = self.to_pixel(dec)
        return dec

########################################################################
# CLIP alignment module
########################################################################
class ClipAlignment(AlignmentModule):
    def __init__(
        self,
        decoder: nn.Module,
        codebook_embed_dim: int,
        img_size: int,
        clip_model_name: str = 'vit_so400m_patch14_siglip_gap_224',
        clip_patch_size: int = 14,
        use_movq: bool = False,
    ):
        super().__init__('clip')
        
        if not TIMM_AVAILABLE:
            raise RuntimeError("timm is required for ClipAlignment. Please install timm.")
        
        # Instantiate the CLIP model
        self.clip_model = create_model(clip_model_name, pretrained=True, img_size=img_size, patch_size=clip_patch_size)
        for p in self.clip_model.parameters():
            p.requires_grad = False
        # Don't set dynamic_img_size=True as it expects spatial format from patch_embed
        # but the model returns flattened tokens (B, L, C)
        self.clip_model.eval()
        
        # Normalization for CLIP
        self.normalize = self._create_normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        self.denormalize = self._create_denormalize([0.5, 0.5, 0.5], [0.5, 0.5, 0.5])
        
        # Use provided decoder
        self.decoder = decoder
        self.post_quant_conv = nn.Linear(codebook_embed_dim, self.decoder.embed_dim)
        self.to_pixel = nn.Linear(self.decoder.embed_dim, self.clip_model.embed_dim)
        self.clip_use_movq = use_movq
    
    def _create_normalize(self, mean, std):
        class Normalize(nn.Module):
            def __init__(self, mean, std):
                super().__init__()
                self.register_buffer('mean', torch.tensor(mean).view(1, -1, 1, 1))
                self.register_buffer('std', torch.tensor(std).view(1, -1, 1, 1))
            def forward(self, x):
                return (x - self.mean) / self.std
        return Normalize(mean, std)
    
    def _create_denormalize(self, mean, std):
        class Denormalize(nn.Module):
            def __init__(self, mean, std):
                super().__init__()
                self.register_buffer('mean', torch.tensor(mean).view(1, -1, 1, 1))
                self.register_buffer('std', torch.tensor(std).view(1, -1, 1, 1))
            def forward(self, x):
                return x * self.std + self.mean
        return Denormalize(mean, std)

    def compute_target(self, x: torch.Tensor) -> torch.Tensor:
        # Preprocess for clip_model: normalize using ImageNet stats
        # First denormalize from [0, 1] to ImageNet range, then normalize
        x_normalized = self.normalize(self.denormalize(x))
        z = self.clip_model.forward_features(x_normalized)[:, self.clip_model.num_prefix_tokens:]
        return z

    def decode_projection(self, quant: torch.Tensor) -> torch.Tensor:
        tmp = quant
        x = self.post_quant_conv(quant)
        if self.clip_use_movq:
            dec = self.decoder(x, interpolate_zq=tmp, H=None, W=None, D=None)
        else:
            dec = self.decoder(x, interpolate_zq=None, H=None, W=None, D=None)
        dec = self.to_pixel(dec)
        return dec


########################################################################
# BiomedCLIP alignment module (mirrors original BiomedClipLoss behavior)
########################################################################
# class BiomedClipAlignment(nn.Module):
#     def __init__(
#         self,
#         embed_dim: int,
#         decoder: nn.Module = None,
#     ):
#         super().__init__()
#         self.embed_dim = embed_dim
#         self.decoder = decoder
#         if self.decoder is None: # then use original default
#             self.decoder = torch.nn.Sequential(
#                 torch.nn.Conv2d(self.embed_dim, 64, 1),
#                 torch.nn.ReLU(),
#                 torch.nn.Conv2d(64, 64, 3, padding="same"),
#                 torch.nn.ReLU(),
#                 torch.nn.Conv2d(64, self.embed_dim, 1),
#             )
#         self.channel_proj = torch.nn.Conv2d(self.embed_dim, self.embed_dim, 1)


#         # Load BiomedCLIP (OpenCLIP) and set to eval/frozen
#         self.clip_model, _, _ = open_clip.create_model_and_transforms(
#             model_name="hf-hub:microsoft/BiomedCLIP-PubMedBERT_256-vit_base_patch16_224",
#         )
#         for p in self.clip_model.parameters():
#             p.requires_grad_(False)
#         self.clip_model.eval()

#         # Preprocessing to mirror the original loss
#         self.transform = T.Compose(
#             [
#                 T.Resize(size=224, interpolation=T.InterpolationMode.BICUBIC, antialias=True),
#                 T.CenterCrop(size=(224, 224)),
#                 T.Normalize(
#                     mean=[0.48145466, 0.4578275, 0.40821073],
#                     std=[0.26862954, 0.26130258, 0.27577711],
#                 ),
#             ]
#         )

#     def forward(self, quant: torch.Tensor, input_image: Optional[torch.Tensor] = None, mask: Optional[torch.Tensor] = None):
#         if input_image is None:
#             raise ValueError("BiomedClipAlignment requires input_image to compute target features")

#         latent = self.channel_proj(self.decoder(quant) + quant)  # (B, L, D)
#         img = self.transform(input_image)
#         img_features = self.clip_model.encode_image(img)  # (B, D)

#         latent = latent / 4.6
#         latent = latent.mean(1, keepdim=True)
#         latent = self.transform(latent.expand(-1, 3, -1, -1))
#         latent_features = self.clip_model.encode_image(latent)

#         img_lat_loss = ((img_features - latent_features) ** 2).sum(1).mean()
#         return img_lat_loss, None


########################################################################
# Vision-Foundation alignment module (VA-VAE style)
########################################################################


class FoundationFeatureExtractor(nn.Module):
    """
    Lightweight wrapper to fetch frozen vision-foundation features.

    Supports MAE and DINOv2-L. Produces spatial feature maps shaped (B, C, H', W').
    """

    def __init__(self, model_type: str):
        super().__init__()
        if not TIMM_AVAILABLE:
            raise RuntimeError("timm is required for FoundationFeatureExtractor.")

        self.model_type = model_type.lower()
        if self.model_type == "mae":
            model_name = "hf-hub:timm/vit_large_patch16_224.mae"
            self.model = create_model(model_name, pretrained=True, dynamic_img_size=True)
            self.patch_size = 16
            self.base_size = 224
            self.feature_dim = 1024
        elif self.model_type == "dinov2":
            model_name = "hf-hub:timm/vit_large_patch14_dinov2.lvd142m"
            self.model = create_model(model_name, pretrained=True, dynamic_img_size=True)
            self.patch_size = 14
            self.base_size = 224
            self.feature_dim = 1024
        elif self.model_type == "biomedclip":
            # OpenCLIP BiomedCLIP ViT-B/16; use encode_image for vision tower
            self.model, _, _ = open_clip.create_model_and_transforms(
                model_name="hf-hub:microsoft/BiomedCLIP-PubMedBERT_256-vit_base_patch16_224",
            )
            self.patch_size = 16
            self.base_size = 224
            # ViT-B hidden dim
            self.feature_dim = 512
            # Normalize inputs like OpenCLIP preprocess
            mean = torch.tensor(self.model.visual.image_mean).view(1, -1, 1, 1)
            std = torch.tensor(self.model.visual.image_std).view(1, -1, 1, 1)
            self.register_buffer("biomed_mean", mean)
            self.register_buffer("biomed_std", std)
        else:
            raise ValueError(f"Unsupported foundation model type: {model_type}")

        self.model.requires_grad_(False)
        self.model.eval()

    @torch.no_grad()
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        b, c, h, w = x.shape
        if self.model_type == "dinov2":
            # DINOv2 expects 224x224 crops; resize then reshape tokens back to spatial grid.
            x_resized = nn.functional.interpolate(x, size=(224, 224), mode='bilinear', align_corners=False)
            tokens = self.model.forward_features(x_resized)[:, 1:]
            # DINOv2 with patch_size=14 on 224x224 gives 16x16 patches
            feat_h = 224 // 14  # = 16
            feat_w = 224 // 14  # = 16
            return tokens.reshape(b, feat_h, feat_w, -1).permute(0, 3, 1, 2)
        if self.model_type == "biomedclip":
            # BiomedCLIP ViT-B/16: use encode_image, returns (B, D)
            x_resized = nn.functional.interpolate(x, size=(self.base_size, self.base_size), mode='bilinear', align_corners=False)
            x_norm = (x_resized - self.biomed_mean.to(x_resized.device)) / self.biomed_std.to(x_resized.device)
            emb = self.model.encode_image(x_norm)  # (B, D)
            return emb.unsqueeze(-1).unsqueeze(-1)  # (B, D, 1, 1)

        # MAE supports dynamic image size; reshape patch tokens (drop cls).
        tokens = self.model.forward_features(x)[:, 1:]
        feat_h = h // self.patch_size
        feat_w = w // self.patch_size
        return tokens.reshape(b, feat_h, feat_w, -1).permute(0, 3, 1, 2)



class VFFoundationAlignment(AlignmentModule):
    """
    Align latent feature maps from the autoencoder with frozen vision foundation
    model features using a two-part VF loss:
      - vf_loss_1: similarity-matrix distance with margin
      - vf_loss_2: per-location cosine margin
    """

    def __init__(
        self,
        latent_channels: int,
        foundation_type: str = "dinov2",
        reverse_proj: bool = True,
        distmat_margin: float = 0.25,
        cos_margin: float = 0.5,
        distmat_weight: float = 1.0,
        cos_weight: float = 1.0,
    ):
        super().__init__('vf')
        self.foundation_model = FoundationFeatureExtractor(foundation_type)
        self.reverse_proj = reverse_proj
        self.distmat_margin = distmat_margin
        self.cos_margin = cos_margin
        self.distmat_weight = distmat_weight
        self.cos_weight = cos_weight

        aux_dim = self.foundation_model.feature_dim
        if reverse_proj:
            # Map latent -> foundation space
            self.linear_proj = nn.Conv2d(latent_channels, aux_dim, kernel_size=1)
        else:
            # Map foundation -> latent space
            self.linear_proj = nn.Conv2d(aux_dim, latent_channels, kernel_size=1)

    def _ensure_4d(self, x: torch.Tensor) -> torch.Tensor:
        if x.dim() == 4:
            return x
        if x.dim() == 3:
            b, l, c = x.shape
            side = int(math.sqrt(l))
            if side * side != l:
                raise ValueError("Latent tokens length is not a perfect square; provide 4D feature maps.")
            return x.transpose(1, 2).reshape(b, c, side, side)
        raise ValueError("Expected latent as (B,C,H,W) or (B,L,C)")

    def _match_spatial(self, a: torch.Tensor, b: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Resize tensors so they share spatial resolution (H, W).
        """
        if a.shape[-2:] == b.shape[-2:]:
            return a, b
        return (
            nn.functional.interpolate(a, size=b.shape[-2:], mode='bilinear', align_corners=False),
            b,
        )

    def compute_target(self, x: torch.Tensor) -> torch.Tensor:
        return self.foundation_model(x)

    def decode_projection(self, quant: torch.Tensor) -> torch.Tensor:
        # For VF alignment, quant is treated as latent feature map (B,C,H,W) or (B,L,C).
        return self._ensure_4d(quant)

    def forward(  # type: ignore[override]
        self,
        quant: torch.Tensor,
        input_image: Optional[torch.Tensor] = None,
        mask: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        if input_image is None:
            raise ValueError("VFFoundationAlignment requires input_image to compute target features.")

        z = self.decode_projection(quant)  # (B, C, H, W)
        aux_feature = self.compute_target(input_image)  # (B, C_aux, H', W')

        # Project to shared channel space
        if self.reverse_proj:
            z_proj = self.linear_proj(z)
            aux_proj = aux_feature
        else:
            aux_proj = self.linear_proj(aux_feature)
            z_proj = z

        # Match spatial shapes
        aux_proj, z_proj = self._match_spatial(aux_proj, z_proj)

        # Compute VF losses
        b, c, h, w = z_proj.shape
        z_flat = z_proj.view(b, c, -1)
        aux_flat = aux_proj.view(b, aux_proj.shape[1], -1)

        z_norm = torch.nn.functional.normalize(z_flat, dim=1)
        aux_norm = torch.nn.functional.normalize(aux_flat, dim=1)

        z_cos_sim = torch.einsum('bci,bcj->bij', z_norm, z_norm)
        aux_cos_sim = torch.einsum('bci,bcj->bij', aux_norm, aux_norm)
        diff = torch.abs(z_cos_sim - aux_cos_sim)

        vf_loss_1 = torch.nn.functional.relu(diff - self.distmat_margin).mean()
        vf_loss_2 = torch.nn.functional.relu(
            1 - self.cos_margin - torch.nn.functional.cosine_similarity(aux_proj, z_proj, dim=1)
        ).mean()

        vf_loss = vf_loss_1 * self.distmat_weight + vf_loss_2 * self.cos_weight
        # print(vf_loss_1, vf_loss_2, vf_loss)
        return vf_loss, z_proj



import torch 
from . import gaussian_diffusion as gd
from .respace import SpacedDiffusion, space_timesteps

def create_gaussian_diffusion(
    *,
    steps=1000,
    learn_sigma=False,
    sigma_small=False,
    noise_schedule="linear",
    use_kl=False,
    predict_xstart=False,
    rescale_timesteps=False,
    rescale_learned_sigmas=False,
    timestep_respacing="",
):
    betas = gd.get_named_beta_schedule(noise_schedule, steps)
    if use_kl:
        loss_type = gd.LossType.RESCALED_KL
    elif rescale_learned_sigmas:
        loss_type = gd.LossType.RESCALED_MSE
    else:
        loss_type = gd.LossType.MSE
    if not timestep_respacing:
        timestep_respacing = [steps]
    return SpacedDiffusion(
        use_timesteps=space_timesteps(steps, timestep_respacing),
        betas=betas,
        model_mean_type=(
            gd.ModelMeanType.EPSILON if not predict_xstart else gd.ModelMeanType.START_X
        ),
        model_var_type=(
            (
                gd.ModelVarType.FIXED_LARGE
                if not sigma_small
                else gd.ModelVarType.FIXED_SMALL
            )
            if not learn_sigma
            else gd.ModelVarType.LEARNED_RANGE
        ),
        loss_type=loss_type,
        rescale_timesteps=rescale_timesteps,
    )